package fi.helsinki.cs.seco.onki.config;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.ResourceFactory;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.vocabulary.*;



public class OnkiConfig {

	private static OnkiConfig configInstance = null;
	private Model configmodel = null;
	private Map<Resource,Configuration> configurations = new HashMap<Resource,Configuration>();
	
	public static final String ONKI_CONF_NS = "http://demo.seco.tkk.fi/onki/2007/07/04/onkiconf#";
	
	/**
	 * Returns the config instance. (singleton)
	 * 
	 * @return
	 */

	public static OnkiConfig getOnkiConfig() {
		if (configInstance == null)
			configInstance = new OnkiConfig();
		return configInstance;
	}
	
	/**
	 * constructor (singleton)
	 */
	
	private OnkiConfig() {
		
	}
	
	/* 
	 * ****************************************************************
	 * Methods for initializing the config model.
	 */

	public void setConfigModel(Model m) {
		this.configmodel = m;
	}
	
	public void readConfigFile(String url) {
		readConfigFile(url, "N3");
	}
	
	public void readConfigFile(String url, String lang) {
		Model m = ModelFactory.createDefaultModel();
		m.read(url, lang);
		if (this.configmodel == null)
			this.configmodel = m;
		else
			this.configmodel.add(m);
	}
	
	/*
	 * Factory methods for creating and getting configuration instances.
	 */
	
	public OnkiServerConfiguration createServerConfiguration() {
		Resource r = this.getSubject(RDF.type, getOnkiConfResource("ServerConfiguration"));		
		Configuration conf = this.getConfigurationInstance(r, OnkiServerConfiguration.class, null);
		if (conf != null)
			return (OnkiServerConfiguration)conf;
		return null;
	}
	
	public DataSourceConfiguration createDataSourceConfiguration(Resource r, Configuration parent) {
		Configuration conf = this.getConfigurationInstance(r, DataSourceConfiguration.class, parent);
		if (conf != null)
			return (DataSourceConfiguration)conf;
		return null;
	}
	
	public LogConfiguration createLogConfiguration(Resource r, Configuration parent) {
		Configuration conf = this.getConfigurationInstance(r, LogConfiguration.class, parent);
		if (conf != null)
			return (LogConfiguration)conf;
		return null;
	}

	public TemplateConfiguration createTemplateConfiguration(Resource r, Configuration parent) {
		Configuration conf = this.getConfigurationInstance(r, TemplateConfiguration.class, parent);
		if (conf != null)
			return (TemplateConfiguration)conf;
		return null;
	}
	
	public ConfigConfiguration createConfigConfiguration(Resource r, Configuration parent) {
		Configuration conf = this.getConfigurationInstance(r, ConfigConfiguration.class, parent);
		if (conf != null)
			return (ConfigConfiguration)conf;
		return null;
	}

	public ServiceConfiguration createServiceConfiguration(Resource r, Configuration parent) {
		Statement stmt = r.getProperty(RDF.type);
		Resource serviceClass = (Resource)stmt.getObject().as(Resource.class);
		
		if (serviceClass.equals(this.getOnkiConfResource("OnkiRedirectService"))) {
			return (ServiceConfiguration)this.getConfigurationInstance(r, OnkiRedirectServiceConfiguration.class, parent);
		}
		if (serviceClass.equals(this.getOnkiConfResource("OntologyService"))) {
			return (OntologyServiceConfiguration)this.getConfigurationInstance(r, OntologyServiceConfiguration.class, parent);
		}
		if (serviceClass.equals(this.getOnkiConfResource("ConceptBrowserService")))
			return (ServiceConfiguration)this.getConfigurationInstance(r, ConceptBrowserServiceConfiguration.class, parent);
		if (serviceClass.equals(this.getOnkiConfResource("MashupQueryService")))
			return (ServiceConfiguration)this.getConfigurationInstance(r, MashupQueryServiceConfiguration.class, parent);
		if (serviceClass.equals(this.getOnkiConfResource("WebserviceQueryService")))
			return (ServiceConfiguration)this.getConfigurationInstance(r, WebserviceQueryServiceConfiguration.class, parent);
		if (serviceClass.equals(this.getOnkiConfResource("GeneralWebserviceService")))
			return (ServiceConfiguration)this.getConfigurationInstance(r, GeneralWebserviceServiceConfiguration.class, parent);
		if (serviceClass.equals(this.getOnkiConfResource("StaticFileDownloadService")))
			return (ServiceConfiguration)this.getConfigurationInstance(r, StaticFileDownloadServiceConfiguration.class, parent);
		if (serviceClass.equals(this.getOnkiConfResource("LuceneSearchService")))
			return (ServiceConfiguration)this.getConfigurationInstance(r, LuceneSearchServiceConfiguration.class, parent);
		if (serviceClass.equals(this.getOnkiConfResource("DocumentationService")))
			return (ServiceConfiguration)this.getConfigurationInstance(r, DocumentationServiceConfiguration.class, parent);
		if (serviceClass.equals(this.getOnkiConfResource("FeedbackService")))
			return (ServiceConfiguration)this.getConfigurationInstance(r, FeedbackServiceConfiguration.class, parent);
		
		
		System.err.println("WARNING: No match in createservieconf for "+r.getLocalName()+"\n");
		return null;
	}
	
	private Configuration getConfigurationInstance(Resource r, Class targetclass, Configuration parent) {
		if (r == null)
			return null;
		if (this.configurations.containsKey(r))
			return this.configurations.get(r);
		
		if (parent != null && r.equals(parent.getClass())) {
			System.err.println("WARNING: Did not create the ConfigurationInstance ("+r+"), because it has the same URI as the parent.");
			return null;
		}
		
		Configuration config = null;
		try {
			Object instance = targetclass.newInstance();
			if (instance == null)
				return null;
			config = (Configuration)instance;
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		}
		
		if (config == null)
			return null;
		
		config.setConfigResource(r);
		config.setOnkiConfig(this);
		config.setParent(parent);
		
		this.configurations.put(r, config);
		return config;
	}
	
	
	/*
	 * ****************************************************************
	 * Helpers etc. 
	 */

	private Resource getSubject(Property predicate, Resource object) {
		StmtIterator i = this.configmodel.listStatements(null, predicate, object);
		if (i.hasNext())
			return i.nextStatement().getSubject();
		return null;
	}
	
	public String toString() {
		String str = "" + this.createServerConfiguration();
		return str;
	}

	public Resource getOnkiConfResource(String localname) {
		return ResourceFactory.createResource(ONKI_CONF_NS+localname);
	}
	
	public Property getOnkiConfProperty(String localname) {
		return ResourceFactory.createProperty(ONKI_CONF_NS+localname);
	}

	public Model getModel() {
		return this.configmodel;
	}
	
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		// TODO Auto-generated method stub
		OnkiConfig conf = OnkiConfig.getOnkiConfig();
		conf.readConfigFile("file:///localhome/kimvilja/workspace/Onki/onki-config.n3");
		System.out.println(conf);
	}

}
