package fi.helsinki.cs.seco.onki.logic;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.math.BigDecimal;
import java.math.MathContext;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.Collection;
import java.util.Collections;

import org.apache.commons.configuration.Configuration;
import org.apache.jcs.JCS;
import org.apache.jcs.access.exception.CacheException;
import org.apache.jcs.engine.stats.behavior.IStatElement;
import org.apache.jcs.engine.stats.behavior.IStats;
import org.apache.log4j.Logger;

import com.hp.hpl.jena.query.QueryExecutionFactory;
import com.hp.hpl.jena.query.QueryFactory;
import com.hp.hpl.jena.query.QuerySolution;
import com.hp.hpl.jena.query.ResultSet;
import com.hp.hpl.jena.rdf.model.HasNoModelException;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.NodeIterator;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Literal;
import com.hp.hpl.jena.rdf.model.ResIterator;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.ResourceFactory;
import com.hp.hpl.jena.rdf.model.Seq;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;
import com.hp.hpl.jena.vocabulary.OWL;
import com.hp.hpl.jena.vocabulary.RDF;
import com.hp.hpl.jena.vocabulary.RDFS;

import fi.helsinki.cs.seco.onki.OnkiException;
import fi.helsinki.cs.seco.onki.service.OnkiDirectory;
import fi.helsinki.cs.seco.onki.service.OnkiDirectoryElement;
import fi.helsinki.cs.seco.onki.service.OnkiHierarchyRelation;
import fi.helsinki.cs.seco.onki.service.OnkiResourceSerializable;
import fi.helsinki.cs.seco.onki.stats.OnkiModelStatistics;
import fi.helsinki.cs.seco.onki.index.AlphabeticalIndex;
import fi.helsinki.cs.seco.onki.index.IndexResult;
import fi.helsinki.cs.seco.onki.index.IndexItem;
import fi.helsinki.cs.seco.onki.index.IndexResults;
import fi.seco.semweb.infrastructure.SecoRegistry;
import fi.seco.semweb.infrastructure.jena.JenaService;

/**
 * Used for processing ontology RDF models.
 * 
 * @author vpkomula
 */

public class OntologyProcessor {

    ////////////////////////////////////////////////////////////
    //	Static variables
    ////////////////////////////////////////////////////////////

public static boolean debugLogEnabled = false;
//public static boolean debugLogEnabled = true;
//public static boolean runTimePrintLogOnlyBlocks = true;
public static boolean runTimePrintLogOnlyBlocks = false;
public static boolean debugToStdOut = false;
//public static boolean debugToStdOut = true;

public static FileOutputStream runTimeLogStream;
public static boolean runTimePrintLog = false;
public static long startRunTime;
public static long prevRunTime;
public static String prevRunTimeId;
public static long parentHierarchyTime;
public static int parentHierarchyCounter;
public static long childHierarchyTime;
public static int childHierarchyCounter;

public static int cacheHits;
public static int cacheMisses;

	// jcs cache
	private Cache uriToLabelCache;
	private static final String uriToLabelCacheRegionName = "uriToLabelCache";

    public static String SERVLET_ROOT_PATH;

    private static OntologyProcessor instance;

	private static final String GROUPS = "groups";
	private static final String TOPCONCEPTS = "topconcepts";
	private static final String ALPHABETICAL = "alphabetical";
	private static final int DEFAULT_DIRECTORY_SIZE = 5;
	
	public static final String SKOS_NS = "http://www.w3.org/2004/02/skos/core#";
	
    ////////////////////////////////////////////////////////////
    //	Private variables
    ////////////////////////////////////////////////////////////

    private Logger logger;

    private Model ontologyModel;
    
    private AlphabeticalIndex ontologyIndex;
    
    private boolean allAttached = false;

    private Configuration config;

    private List labelProps;

    private long filesLastModified;
    
    private Date startTime;
    
    private Date cacheExpireTime;
    
    private HashMap<String, HashMap<String, OnkiDirectory>> directories =
    					new HashMap<String, HashMap<String, OnkiDirectory>>();

    ////////////////////////////////////////////////////////////
    //	Constructors
    ////////////////////////////////////////////////////////////

    /** 
     * Singleton class, prevent instantation.
     */
    
    protected OntologyProcessor() {
        ;
    }

    private OntologyProcessor(Configuration configuration) throws Exception {
        this.logger = Logger.getLogger(getClass());
        this.config = configuration;
        this.uriToLabelCache = new Cache(uriToLabelCacheRegionName);
    }

    public static OntologyProcessor getInstance() {
        return instance;
    }
    
    public static OntologyProcessor getInstance(Configuration conf) throws Exception {
        if (instance == null) {
            if (conf == null) {
                throw new NullPointerException("Parameter given configuration is null");
            }
            instance = new OntologyProcessor(conf);
        }
        return instance;
    }

    ////////////////////////////////////////////////////////////
    //	Methods for accessing configuration etc.
    ////////////////////////////////////////////////////////////
    
    /**
	 * Get the Apache Commons configuration.
	 */
    
    public Configuration getConfig() {
        return this.config;
    }
    
    /**
	 * Get the logger.
	 */
    
    public Logger getLogger() {
    	return this.logger;
    }
    
    /**
	 * Get certain property types (defined in the configuration) as a list.
	 */
    
    public List<String> getPropsAsList(int propertyType) {
    	String[] props = this.getProps(propertyType);
    	if (props != null)
    		return Arrays.asList(props);
    	return null;
    }
    
    public String[] getProps(int propertyType) {
    	if (propertyType == OnkiResource.LABEL)
    		return this.config.getStringArray(OnkiConfigurator.ELEM_CLASSES_LABEL);
    	else if (propertyType == OnkiResource.ALT_LABEL)
    		return this.config.getStringArray(OnkiConfigurator.ELEM_CLASSES_INDEX_ALT_LABEL);
    	else if (propertyType == OnkiResource.DESCRIPTION)
    		return this.config.getStringArray(OnkiConfigurator.ELEM_CLASSES_DESCRIPTION_PROPERTIES_PROPERTY);
    	else if (propertyType == OnkiResource.RELATED_CONCEPT)
    		return this.config.getStringArray(OnkiConfigurator.ELEM_CLASSES_RELATED_PROPERTIES_PROPERTY);
    	else if (propertyType == OnkiResource.EQUIVALENT_CONCEPT)
    		return this.config.getStringArray(OnkiConfigurator.ELEM_CLASSES_EQUIVALENT_CLASS_PROPERTIES_PROPERTY);
    	return null;
    }
    
    /**
	 * Get the properties for labels.
	 */
    
    public List getLabelProps() {
    	if (this.labelProps == null || this.labelProps.isEmpty()) {
	        //	labels
	        String[] labels = this.config.getStringArray(OnkiConfigurator.ELEM_CLASSES_LABEL);
	        this.labelProps = new ArrayList();
	        for (int i = 0; i < labels.length; i++) {
	            String labelPredStr = labels[i];
	            if (labelPredStr != null) {
	                Property p = ontologyModel.getProperty(labelPredStr);
	                if (p != null) {
	
	                    this.labelProps.add(p);
	                }
	            }
	        }
    	}
        return this.labelProps;
    }
    
    /**
	 * Get the properties for concept hierarchy.
	 */
    
    public String getFollowHierarchyProp() {
    	return config.getString(OnkiConfigurator.ELEM_CLASSES_FOLLOW);
    }
    
    /**
	 * Get the properties marked as "additional" for UI.
	 */
    
    public String[] getAdditionalProps() {
    	String[] additionals = config.getStringArray(OnkiConfigurator.ELEM_CLASSES_ADDITIONAL);
    	return additionals;
    }
    
    /**
	 * Get the properties to be hidden in the UI.
	 */
    
    public String[] getHiddenProps() {
    	String[] hiddenProps = config.getStringArray(OnkiConfigurator.ELEM_CLASSES_HIDDEN_PROPERTY);
    	return hiddenProps;
    }
    
    /**
	 * Get the properties for "related" relations.
	 */
    
    public String[] getRelatedProps() {
    	String[] relatedProps = config.getStringArray(OnkiConfigurator.ELEM_CLASSES_RELATED_PROPERTIES_PROPERTY);
    	return relatedProps;
    }
    
    /**
	 * Get the properties for textual descriptions.
	 */
    
    public String[] getDescriptionProps() {
    	String[] descriptionProps = config.getStringArray(OnkiConfigurator.ELEM_CLASSES_DESCRIPTION_PROPERTIES_PROPERTY);
    	return descriptionProps;
    }
    
    /**
	 * Load other configuration on instantiation.
	 */
    
    private void loadOtherConfiguration() throws Exception {
    	OnkiResource.initialize(this);
    }
    
    /**
	 * Get statistics about the ontology.
	 */
    
    public OnkiModelStatistics getOntologyStatistics() {
        return OnkiModelStatistics.getInstance(ontologyModel);
    }

    /**
	 * Get the alphabets for the alphabetical index.
	 */

    public List getAlphabets() {
    	return config.getList(OnkiConfigurator.ELEM_CLASSES_ALPHA_INDEX);
    }

    /**
	 * Set the start time of the application (for REST).
	 */
    
    public void setStartTime(Date startTime) {
    	this.startTime = startTime;
    }
    
    /**
	 * Get the start time of the application.
	 */
    
    public Date getStartTime() {
    	return this.startTime;
    }
    
    /**
	 * Set the cache expire time (for REST).
	 */
    
    public void setCacheExpireTime(Date cacheExpireTime) {
    	this.cacheExpireTime = cacheExpireTime;
    }
    
    /**
	 * Get the cache expire time.
	 */
    
    public Date getCacheExpireTime() {
    	return this.cacheExpireTime;
    }
    
    /**
	 * Get the default type restriction.
	 */
    
    public Collection<String> getDefaultViewTypes() {
    	return config.getList(OnkiConfigurator.ELEM_DEFAULTVIEW_TYPE);
    }
    
    /**
	 * Get the languages of labels the concept index contains. 
	 */
    
    public Collection<String> getAvailableLanguages() {
    	return this.ontologyIndex.getKnownLangs();
    }
    
    /**
	 * Get the alphabets (for alphabetical index) that concept index contains.
	 */
    
    public Collection<Character> getAvailableAlphabets(String lang) {
    	LinkedList<Character> list = new LinkedList<Character>();
    	Collection<Character> alphabetsInIndex = this.ontologyIndex.getKnownAlphabets(lang);
    	if (alphabetsInIndex != null) {
	    	for (char alphabet : alphabetsInIndex) {
	    		// let's not add alphabets which give 0 hits with a given language
	    		IndexResults results = this.searchConcept(String.valueOf(alphabet), lang, null, null, null, false, false, -1, -1);
	    		if (results.getMetadata().getTotalHitsAmount() > 0)
	    			list.add(alphabet);
	    	}
    	}
    	return list;
    }    
    
    /**
	 * Get the label prefixes that concept index contains.
	 */
    
    public Map<String, Integer> getAvailableLabelPrefixes(String lang) {
    	return this.ontologyIndex.getKnownLabelPrefixes(lang);
    	// searching index is too slow
    	/*Map<String, Integer> map = new TreeMap<String, Integer>();
    	Collection<String> labelPrefixesInIndex = this.ontologyIndex.getKnownLabelPrefixes(lang).keySet();
    	if (labelPrefixesInIndex != null) {
	    	for (String labelPrefix : labelPrefixesInIndex) {
	    		// let's not add label prefixes which give 0 hits with a given language
	    		IndexResults results = this.searchConcept(labelPrefix, lang, null, null, null, false, false, -1, -1);
	    		if (results.getMetadata().getTotalHitsAmount() > 0)
	    			map.put(labelPrefix, results.getMetadata().getTotalHitsAmount());
	    	}
    	}
    	return map;*/
    }    
    
    /**
	 * Get the types of concepts the concept index contains.
	 */
    
    public Collection<OnkiResource> getAvailableTypes(List langPreferences) {
    	List<OnkiResource> types = new LinkedList<OnkiResource>(); 
    	for (String uri : this.ontologyIndex.getKnownTypes()) 
    		types.add(this.createOnkiResource(uri, langPreferences));
    	Collections.sort(types);
    	return types;
    }
    
    public Collection<String> getAvailableTypeUris() {
    	return this.ontologyIndex.getKnownTypes();
    }
    
    /**
	 * Get the namespaces of concepts the concept index contains.
	 */
    
    public Collection<String> getAvailableConceptNamespaces() {
    	return this.ontologyIndex.getKnownNamespaces();
    }
    
    /**
	 * Get the language preference order.
	 */
    
    public List getLangPreferences(String lang) {
    	List langPreferences = new LinkedList(this.getConfig().getList(OnkiConfigurator.ELEM_LANGUAGES_PREFERENCE));
		if (lang != null)
			((LinkedList)langPreferences).addFirst(lang);
        langPreferences.add(null);
    	return langPreferences;
    }
    
    ////////////////////////////////////////////////////////////
    //	Methods for accessing Ontology repository
    ////////////////////////////////////////////////////////////

    /**
	 * Get the source URLs of the ontologies.
	 */
    
    public Collection<String> getOntologySourceUrls() {
    	JenaService js = SecoRegistry.getService("JenaService",JenaService.class);
    	Set<String> ontologySourceUrls = new HashSet<String>();
    	for (String sourceUrl : js.getSourceUrls("onkiModel")) {
    		ontologySourceUrls.add(sourceUrl);
    		// return also possible SKOS versions of the ontology files
    		// (in the same dir, file names matching pattern "*-skos*")
    		if (sourceUrl.startsWith("file:/"))
	    		for (File siblingFile : new File(sourceUrl.substring(5)).getParentFile().listFiles())
	    			if (siblingFile.getName().contains("-skos."))
	    				ontologySourceUrls.add("file:"+siblingFile.getPath());
    	}
    	System.out.println(ontologySourceUrls);
    	return ontologySourceUrls;
    }
    
    /**
	 * Get the last modified timestamp of the ontology sources.
	 */
    
    public long getFilesLastModified() {
    	return this.filesLastModified;
    }
    
    /**
     * Load the ontologies to a Jena RDF model.
     * 
     * @throws IOException
     * @throws OnkiException
     */
    
    public synchronized void loadOntologies() throws IOException, OnkiException, Exception {
        long startTime = System.currentTimeMillis();
        
        JenaService js = SecoRegistry.getService("JenaService",JenaService.class);
        LinkedList<String> ignoreKeys = new LinkedList<String>();
        ignoreKeys.add("visualization");
        this.filesLastModified = js.getFileLastModified("onkiModel", ignoreKeys);
        ontologyModel = js.getModel("onkiModel");

        if (allAttached)
            return;

        allAttached = true;
        long endTime = System.currentTimeMillis();
        logger.info("loadOntologies() Ontologies loaded in " + ((double) (endTime - startTime)) / 1000.0 + " secs.");
        
        loadOtherConfiguration();
    }
    
    /**
	 * Does the RDF model contain a specific resource.
	 */
    
    public boolean containsResource(Resource res) {
    	return ontologyModel.containsResource(res);
    }
    
    public boolean containsResource(String uri) {
		if (uri == null)
		    return false;
		uri = uri.trim();
		if( uri.length()==0)
			return false;
		Resource res = ResourceFactory.createResource(uri);
		return containsResource(res);
    }
    
    /**
	 * Create OnkiResource object.
	 */
    
    public OnkiResource createOnkiResource(String uri, List langPreferences) {
        Resource current = ontologyModel.getResource(uri);
        if (current != null)
        	return new OnkiResource(current, langPreferences);
        return null;
    }
    
    /**
     * Get concept as RDF resource.
     */
    
    public Resource getConcept(String uri) {
        if (!this.containsResource(uri))
            return null;
        return ontologyModel.getResource(uri);
    }
    
    /**
	 * Gets the first ontology of the model as a ontology resource.
     * Used to get metadata about ontology.
	 */

    public Resource getOntologyResource()  {
    	//TODO: support for more than one ontologies (MultiOnki)
    	for (ResIterator iter = ontologyModel.listSubjectsWithProperty(RDF.type, OWL.Ontology); iter.hasNext(); )
    		return iter.nextResource();
    	return null;
    }
    
    /**
	 * Get the concept to be shown in the default view (front page of the browser).
	 */
    
    public String getDefaultViewConcept() {
    	String defaultViewConcept = config.getString(OnkiConfigurator.ELEM_DEFAULTVIEW_CONCEPT);
    	if (defaultViewConcept != null && defaultViewConcept != "")
    		return defaultViewConcept;
    	// if defaultViewConcept isn't defined in conf, get first ConceptScheme
    	Iterator<Resource> schemeIter = this.getConceptSchemes().iterator();
    	if (schemeIter.hasNext())
    		return schemeIter.next().getURI();
    	return null;
    }
    
    /**
	 * Generate the concept hierarchy of a concept.
	 */
    
    public void addConceptHierarchy(OnkiResource ores, List langPreferences) {
        addParentConcepts(ores, langPreferences);
        addSiblingConcepts(ores, langPreferences);
        addChildConcepts(ores, langPreferences);
    }
    
    /**
	 * Generate the parents of a concept.
	 */
    
    public void addParentConcepts(OnkiResource ores, List langPreferences, int depth) {
    	ores.setParents(null);
    	ores.setRoots(null);
    	// support for multiple hierarchy properties (e.g. skos:broader, skos:narrower)
    	Map<Property,String> hierarchyProperties = new HashMap<Property,String>();
    	String[] follow = config.getStringArray(OnkiConfigurator.ELEM_CLASSES_FOLLOW);
    	for (int i=0; i<follow.length; i++) {
    		Property followHierarchyProp = ontologyModel.getProperty(follow[i]);
    		String type = OnkiConfigurator.getType(OnkiConfigurator.ELEM_CLASSES_FOLLOW, i);
    		hierarchyProperties.put(followHierarchyProp, type);
    	}
    	
		Map<Resource,OnkiResource> visitedNodes = new HashMap<Resource,OnkiResource>();
		getParentClasses2(ores, hierarchyProperties, visitedNodes, langPreferences, depth);
    }
    
    public void addParentConcepts(OnkiResource ores, List langPreferences) {
    	addParentConcepts(ores, langPreferences, -1);
    }
    
    private Set<OnkiResource> getParentClasses2(OnkiResource current, 
    		Map<Property,String> hierarchyProperties, 
    		Map<Resource,OnkiResource> visitedNodes, 
    		List langPreferences, int depth) {

    	visitedNodes.put((Resource)current.getResource(), current);

    	Set<OnkiResource> roots = new HashSet<OnkiResource>();

    	// support for skos:ConceptScheme
    	//TODO: add support for equivalentClass for topConcepts
    	if (!current.isConceptScheme()) {
    		
    		//add conceptSchemes as parents
	    	if (current.isTopConcept()) {
	    		Iterator<Resource> resIter = this.getConceptSchemes((Resource)current.getResource()).iterator();
	    		Set<OnkiResource> parentRoots = this.getParentClasses3(current, hierarchyProperties, visitedNodes, langPreferences, resIter, depth);
				roots.addAll(parentRoots);
	    	}
	    	    	
	    	// support for skos:Collection as node label
	    	else if (config.getBoolean(OnkiConfigurator.ELEM_CLASSES_SKOS_COLLECTION_AS_NODE_LABEL_ENABLED, false) && current.isCollection()) {
				// parents of a collection are the collections this collection is a member of 
				Set<OnkiResource> parents = getCollections(current, langPreferences);
				// or the intersection of all the parent (or nearest ancestor) concepts of the (sub)collection(s) members
				if (parents.isEmpty())
					parents = this.getParentConcepts(current, hierarchyProperties, langPreferences);
				
				for (OnkiResource parent : parents) {
					current.addParent(parent);
					Set<OnkiResource> parentRoots = this.getParentClasses2(parent, hierarchyProperties, visitedNodes, langPreferences, depth);
					roots.addAll(parentRoots);
				}
	    	}
	    	
	    	else {
	    		for (Iterator<Map.Entry<Property,String>> i = hierarchyProperties.entrySet().iterator(); i.hasNext(); ) {
	    			Map.Entry<Property,String> entry = i.next();

				    // add direct parents
				    
				    // List all parents (based on the current hierarchy property)
	    			ExtendedIterator iter;
long start = System.nanoTime();
	    			if (entry.getValue().equals("subClassOf"))
	    				iter = this.ontologyModel.listObjectsOfProperty((Resource)current.getResource(), entry.getKey());
	    			else
	    				iter = this.ontologyModel.listSubjectsWithProperty(entry.getKey(), current.getResource());
parentHierarchyTime += System.nanoTime()-start;
parentHierarchyCounter++;
				    Set<OnkiResource> parentRoots = this.getParentClasses3(current, hierarchyProperties, visitedNodes, langPreferences, iter, depth);
					roots.addAll(parentRoots);
					
					// add parents of equivalent concepts
				    if (config.getBoolean(OnkiConfigurator.ELEM_MULTIONKI_ENABLED) && !config.getBoolean(OnkiConfigurator.ELEM_MULTIONKI_NOMAPPINGSINHIERARCHY, false)) {
				    	
				    	for (Resource equivResource : this.getEquivalentClassesRes(current.getResource(), null)) {

				    		// list all parents of equivalent class (based on the current hierarchy property)

				    		ExtendedIterator equivI;
				    		if (entry.getValue().equals("subClassOf"))
				    			equivI = this.ontologyModel.listObjectsOfProperty(equivResource, entry.getKey());
				    			
				    		else
				    			equivI = this.ontologyModel.listSubjectsWithProperty(entry.getKey(), equivResource);
				    			
				    		parentRoots = this.getParentClasses3(current, hierarchyProperties, visitedNodes, langPreferences, equivI, depth);
				    		
				    		// if parentRoots are skos:Collections, do not add them here
				    		if ((!parentRoots.isEmpty() && !parentRoots.iterator().next().isCollection()))
				    			roots.addAll(parentRoots);
				    	}
				    }
				}
	    	}
    	}
		
		// if no roots found, this should be a root (conceptScheme)
    	if (roots.isEmpty()) {
			roots = new HashSet<OnkiResource>();
			roots.add(current);
		}

    	// eliminate the resource itself from being its root, if there's another root(s)
    	if (roots.contains(current) && roots.size() > 1)
			roots.remove(current);
    	
    	if (depth < 0) // unrestricted transitive parent search
    		current.addRoots(roots);
		return roots;
    }
    
    private Set<OnkiResource> getParentClasses3(OnkiResource current, Map<Property,String> hierarchyProperties,
            Map<Resource,OnkiResource> visitedNodes,
            List langPreferences, Iterator iter, int depth) {

    	depth--;
    	
    	Set<OnkiResource> roots = null;
    	
    	// support for skos:Collection as node label
    	Collection<OnkiResource> parentsOfParentCollections = new HashSet<OnkiResource>();
    	if (config.getBoolean(OnkiConfigurator.ELEM_CLASSES_SKOS_COLLECTION_AS_NODE_LABEL_ENABLED, false)) {
    		// loop through collections current is member of
    		for (OnkiResource ores : this.getCollections(current, langPreferences)) {
    			Resource res = ores.getResource();
	
    			if (res == null || res.equals(RDFS.Resource) || res.isAnon())
    				continue;

    			Set<OnkiResource> parentRoots = null;
	
    			if (visitedNodes.containsKey(res))
    				ores = visitedNodes.get(res);
	
				if (ores == current)
					continue;
				
				current.addParent(ores);
				
				if (visitedNodes.containsKey(res)) {
					ores.addSubnodesRecursively(current.getSubnodes());
					ores.addSubnodeRecursively(current);				
					parentRoots = ores.getRoots();
				} else if (depth != 0)
					parentRoots = getParentClasses2(ores, hierarchyProperties, visitedNodes, langPreferences, depth);

				if (parentRoots != null) {
					if (roots == null)
						roots = new HashSet<OnkiResource>();
					roots.addAll(parentRoots);
				}
				parentsOfParentCollections.addAll(ores.getParents());
			}
    	}
    	
        // "iter" contains parent concepts of the "current" 	
	    while (iter.hasNext()) {
	    		
	    	Resource res = null;
	    	res = (Resource)iter.next();
	    	
			if (res == null || res.equals(RDFS.Resource) || res.isAnon())
				continue;
			
			OnkiResource ores;
			Set<OnkiResource> parentRoots = null;
			
			if (visitedNodes.containsKey(res)) {
				ores = visitedNodes.get(res);
			} else {
				ores = new OnkiResource(res, langPreferences);					
			}
			
			if (ores == current)
				continue;
			
	    	// do not add parent concepts that are shared with all the collection members
			if (parentsOfParentCollections.contains(ores))
				continue;
			
			// add parent
			current.addParent(ores);
			
			if (config.getBoolean(OnkiConfigurator.ELEM_MULTIONKI_ENABLED) && !config.getBoolean(OnkiConfigurator.ELEM_MULTIONKI_NOMAPPINGSINHIERARCHY, false)) {
			// loop through equivalent concepts of current and parent
			
				for (Resource equivParentRes : ores.getEquivalentClassesRes()) {
					OnkiResource equivParentOres;
					if (visitedNodes.containsKey(equivParentRes))
						equivParentOres = visitedNodes.get(equivParentRes);
					else
						equivParentOres = new OnkiResource(equivParentRes, langPreferences);
					
					equivParentOres.addChild(current);
					
					for (Resource equivRes : current.getEquivalentClassesRes()) {
						OnkiResource equivOres;
						if (visitedNodes.containsKey(equivRes))
							equivOres = visitedNodes.get(equivRes);
						else
							equivOres = new OnkiResource(equivRes, langPreferences);
						
						equivOres.addParent(ores);
					}
				}
			}
		
			//should these be removed if ores is not added as current's parent (multionki)?
			if (visitedNodes.containsKey(res)) {
				ores.addSubnodesRecursively(current.getSubnodes());
				ores.addSubnodeRecursively(current);					
				parentRoots = ores.getRoots();
			} else if (depth != 0)
				parentRoots = getParentClasses2(ores, hierarchyProperties, visitedNodes, langPreferences, depth);

			if (parentRoots != null) {
				if (roots == null)
					roots = new HashSet<OnkiResource>();
				roots.addAll(parentRoots);
			}/* else {
				System.out.println("WARNING: We're getting null roots at "+current);
			}*/
	    }
    	
    	// if no roots found, this should be a root
    	if (roots == null) {
			roots = new HashSet<OnkiResource>();
			roots.add(current);
		}
    	
    	return roots;
    }

    /**
     * Get the parent concepts of a concept/collection without generating the hierarchy tree.
     */
    
    public Set<OnkiResource> getParentConcepts(OnkiResource current, Map<Property,String> hierarchyProperties, List langPreferences) {
    	if (current.isCollection())
    		return getParentConceptsOfCollection(current, hierarchyProperties, langPreferences);
    	else
    		return getParentConceptsOfConcept(current, hierarchyProperties, langPreferences);
    }
    
    /**
     * Get the parent concepts of a concept without generating the hierarchy tree.
     */
    
    private Set<OnkiResource> getParentConceptsOfConcept(OnkiResource current, Map<Property,String> hierarchyProperties, List langPreferences) {
    	Set<OnkiResource> parents = new HashSet<OnkiResource>();
    	for (Iterator<Map.Entry<Property,String>> i = hierarchyProperties.entrySet().iterator(); i.hasNext(); ) {
    		Map.Entry<Property,String> entry = i.next();

    		ExtendedIterator iter;

    		if (entry.getValue().equals("subClassOf"))
    			iter = this.ontologyModel.listObjectsOfProperty((Resource)current.getResource(), entry.getKey());
    		else
    			iter = this.ontologyModel.listSubjectsWithProperty(entry.getKey(), current.getResource());
    		
    		while (iter.hasNext()) {
	    		Resource res = null;
	    		res = (Resource)iter.next();
	
				if (res == null || res.equals(RDFS.Resource) || res.isAnon())
				    continue;

				parents.add(new OnkiResource(res, langPreferences));
    		}
    	}
    	return parents;
    }
    
	/**
	 * Get the parent concepts of a collection without generating the hierarchy tree.
	 */
	
	private Set<OnkiResource> getParentConceptsOfCollection(OnkiResource collection, Map<Property,String> hierarchyProperties, List langPreferences) {
		Set<OnkiResource> parents = new HashSet<OnkiResource>();
		Set<OnkiResource> members = this.getCollectionMembers(collection, langPreferences, true);
		
		for (OnkiResource member : members) {
			if (member.isCollection())
				continue;

			// get the parent concepts of a member concept and see if the other members have these concepts as ancestors
			// if yes, add them as a shared parent
			// if not, traverse the concept hierarchy up until shared ancestors are found
			Set<OnkiResource> parentCandidates = this.getParentConcepts(member, hierarchyProperties, langPreferences);
			while (!parentCandidates.isEmpty()) {
				// prevent loops (due to data bugs, if there are triples <concept> <skos:broaderGeneric> <collection> and <collection> <skos:member> <concept>)
				if (parentCandidates.contains(collection))
					parentCandidates.remove(collection);
				for (OnkiResource parent : parentCandidates) {
					boolean sharedParent = true;
					for (OnkiResource member2 : members) {
						if (member2.equals(member))
							continue;
						if (member2.isCollection())
							continue;
						if (this.getAncestorConcepts(member2, hierarchyProperties, langPreferences).contains(parent))
							continue;
						sharedParent = false;
						break;
					}
					if (sharedParent)
						parents.add(parent);
				}
				parentCandidates.removeAll(parents);
				Set<OnkiResource> tempParentCandidates = new HashSet<OnkiResource>();
				for (OnkiResource parent : parentCandidates)
					tempParentCandidates.addAll(this.getParentConcepts(parent, hierarchyProperties, langPreferences));
				parentCandidates = tempParentCandidates;
			}

			break; // only one member concept needs to be processed
		}
		return parents;
	}
    
    /**
     * Get the ancestor concepts of a concept without generating the hierarchy tree.
     */

    public Set<OnkiResource> getAncestorConcepts(OnkiResource current, Map<Property,String> hierarchyProperties, List langPreferences) {
    	Set<OnkiResource> ancestors = new HashSet<OnkiResource>();
    	for (OnkiResource parent : getParentConcepts(current, hierarchyProperties, langPreferences)) {
    		if (!parent.isCollection()) {
    			ancestors.add(parent);
    			ancestors.addAll(this.getAncestorConcepts(parent, hierarchyProperties, langPreferences));
    		}
    	}
    	return ancestors;
    }
    
    /**
     * Get the collections (parents) a concept is a member of.
     */
    
    public Set<OnkiResource> getCollections(OnkiResource current, List langPreferences) {
    	Set<OnkiResource> collections = new HashSet<OnkiResource>();
		Property member = ResourceFactory.createProperty(SKOS_NS+"member");
		for (StmtIterator iter = this.ontologyModel.listStatements(null, member, current.getResource()); iter.hasNext(); )
			collections.add(new OnkiResource(iter.nextStatement().getSubject(), langPreferences));
		return collections;
    }
    
    /**
	 * Generate the children of a concept.
	 */
    
    public void addChildConcepts(OnkiResource ores, List langPreferences) {
    	printRunTime("   siblingTime", "8.1");
    	// support for multiple hierarchy properties (e.g. skos:broader, skos:narrower) 
    	Map<Property,String> hierarchyProperties = new HashMap<Property,String>();
    	String[] follow = config.getStringArray(OnkiConfigurator.ELEM_CLASSES_FOLLOW);
    	printRunTime("   siblingTime", "8.2");
    	for (int i=0; i<follow.length; i++) {
    		printRunTime("   siblingTime", "8.3");
    		Property followHierarchyProp = ontologyModel.getProperty(follow[i]);
    		printRunTime("   siblingTime", "8.4");
    		String type = OnkiConfigurator.getType(OnkiConfigurator.ELEM_CLASSES_FOLLOW, i);
    		printRunTime("   siblingTime", "8.5");
    		hierarchyProperties.put(followHierarchyProp, type);
    	}
    	printRunTime("siblingTime", "8");
		this.getChildConcepts2(ores, hierarchyProperties, langPreferences);
    }
    
    private void getChildConcepts2(OnkiResource current, 
    		Map<Property,String> hierarchyProperties, 
    		List langPreferences) {
    	//support for skos:ConceptScheme
    	if (current.isConceptScheme()) {
    		Resource topConcept = ResourceFactory.createResource(SKOS_NS+"TopConcept");
    		StmtIterator iter = ontologyModel.listStatements(null, RDF.type, topConcept);
    		if (iter != null)
    			while (iter.hasNext()) {
    				OnkiResource child = new OnkiResource(iter.nextStatement().getSubject(), langPreferences);
    				current.addChild(child);
    			}
    		Property hasTopConcept = ResourceFactory.createProperty(SKOS_NS+"hasTopConcept");
    		StmtIterator iter2 = ontologyModel.listStatements((Resource)current.getResource(), hasTopConcept, (RDFNode)null);
    		if (iter2 != null)
    			while (iter2.hasNext()) {
    				OnkiResource child = new OnkiResource((Resource)iter2.nextStatement().getObject(), langPreferences);
    				current.addChild(child);
    			}
    	}
    	// support for skos:Collection as node label
    	else if (config.getBoolean(OnkiConfigurator.ELEM_CLASSES_SKOS_COLLECTION_AS_NODE_LABEL_ENABLED, false) && current.isCollection()) {
    		for (OnkiResource member : this.getCollectionMembers(current, langPreferences, false))
				current.addChild(member);
    	}
    	else {
    		Collection<OnkiResource> children = new HashSet<OnkiResource>();
    		printRunTime("siblingTime", "9");
    		for (Iterator<Map.Entry<Property,String>> i = hierarchyProperties.entrySet().iterator(); i.hasNext(); ) {
    			Map.Entry<Property,String> entry = i.next();
			    
    			ExtendedIterator resI;
long start = System.nanoTime();
    			if (entry.getValue().equals("subClassOf"))
    				resI = ontologyModel.listSubjectsWithProperty(entry.getKey(), current.getResource());
    			else
    				resI = ontologyModel.listObjectsOfProperty(current.getResource(), entry.getKey());
childHierarchyTime += System.nanoTime()-start;
childHierarchyCounter++;
			    while (resI.hasNext()) {
			    	Resource childRes = (Resource)resI.next();
			    	if (!childRes.isAnon()) {
				    	OnkiResource child = new OnkiResource(childRes, langPreferences);
				    	printRunTime("siblingTime", "10");
				    	children.add(child);
				    	printRunTime("siblingTime", "11");
			    	}
			    	//child.addTreeId(current); TODO: Fix this ???
			    }
			    if (config.getBoolean(OnkiConfigurator.ELEM_MULTIONKI_ENABLED) && !config.getBoolean(OnkiConfigurator.ELEM_MULTIONKI_NOMAPPINGSINHIERARCHY, false)) {
			    	for (Resource equivResource : this.getEquivalentClassesRes(current.getResource(), null)) {
			    		ExtendedIterator resEquivChildsI;
			    		if (entry.getValue().equals("subClassOf"))
			    			resEquivChildsI = ontologyModel.listSubjectsWithProperty(entry.getKey(), equivResource);
		    			else
		    				resEquivChildsI = ontologyModel.listObjectsOfProperty(equivResource, entry.getKey());
			    		
			    		while (resEquivChildsI.hasNext()) {
			    			OnkiResource equivChild = new OnkiResource((Resource)resEquivChildsI.next(), langPreferences);
			    			children.add(equivChild);
			    		}
			    	}
			    }
			}
    		
        	// support for skos:Collection as node label
    		if (config.getBoolean(OnkiConfigurator.ELEM_CLASSES_SKOS_COLLECTION_AS_NODE_LABEL_ENABLED, false)) {
	    		// get the skos:Collections child concepts are member of
	    		// only if all members are children (or descendants) of this concept 
	    		Collection<OnkiResource> childCollections = new HashSet<OnkiResource>();
	    		for (OnkiResource child : children) {
	    			for (OnkiResource collection : this.getCollections(child, langPreferences)) {
	    				boolean addAsChild = true;
	    				for (OnkiResource member : this.getCollectionMembers(collection, langPreferences, true)) {
	    					if (!member.isCollection() && !children.contains(member)) {
	    						Set<OnkiResource> memberAncestors = this.getAncestorConcepts(member, hierarchyProperties, langPreferences);
	    						memberAncestors.retainAll(children);
	    						if (memberAncestors.isEmpty()) {
	    							addAsChild = false;
	    							break;
	    						}
	    					}
	    				}
	    				if (addAsChild)
	    					childCollections.add(collection);
	    			}
	    		}
	    		// remove those concepts as children of current concept
	    		for (OnkiResource collection : childCollections) {
	    			Set<OnkiResource> members = getCollectionMembers(collection, langPreferences, true);
	    			children.removeAll(members);
	    			// add collections as children
	    			children.add(collection);
	    		}
    		}
    		for (OnkiResource child : children)
    			current.addChild(child);
    	}
    }

    /**
     * Get the members of a collection.
     */
    
    public Set<OnkiResource> getCollectionMembers(OnkiResource collection, List langPreferences, boolean transitive) {
		Set<OnkiResource> members = new HashSet<OnkiResource>();
		Property member = ResourceFactory.createProperty(SKOS_NS+"member");
		for (StmtIterator iter = collection.getProperties(member); iter.hasNext(); ) {
			OnkiResource memberOres = new OnkiResource(iter.nextStatement().getResource(), langPreferences);
			members.add(memberOres);
			if (!members.contains(memberOres)) {
				if (transitive && memberOres.isCollection())
					members.addAll(this.getCollectionMembers(memberOres, langPreferences, transitive));
			}
		}
		return members;
	}
    
    /**
	 * Generate the siblings of a concept.
	 */
    
    public void addSiblingConcepts(OnkiResource ores, List langPreferences) {
    	printRunTime("siblingTime", "4");
        Collection<OnkiResource> parents = ores.getParents();
        printRunTime("siblingTime", "5");
        for (OnkiResource parent: parents) {
        	printRunTime("siblingTime", "6");
        	this.addChildConcepts(parent, langPreferences);
        }     
        printRunTime("siblingTime", "7");
    }

    /**
	 * Get the roots of the concept hierarchy of a concept.
	 */
    
    public Collection<OnkiResource> getHierarchyRoots(String uri, List<String> langPreferences) {
    	if (uri != null) {
    		OnkiResource ores = this.createOnkiResource(uri, langPreferences);
    		this.addConceptHierarchy(ores, langPreferences);
    		return ores.getRoots();
    	}
    	// TODO: could be HashSet of OnkiResources instead of two structures
    	LinkedList<OnkiResource> rootResources = new LinkedList<OnkiResource>();
    	for (IndexResult result : this.searchConcept("*", null, null, null, null, false, true, -1, -1).getResults())
    		rootResources.add(this.createOnkiResource(result.getUri(), langPreferences));
   		return rootResources;
    }
    
    /**
	 * Get the concept tree of a concept.
	 */
    
    public Collection<OnkiHierarchyRelation> getConceptTree(String uri, List langPreferences) {
    	return this.getConceptTree(uri, langPreferences, -1, 0, 0);
    }
    
    public Collection<OnkiHierarchyRelation> getConceptTree(String uri, List langPreferences, int parents, int children, int siblings) {
    	// TODO: implement parents == n, "full tree" (uri == null)
    	if (this.createOnkiResource(uri, null) != null) {
    	
	    	LinkedList<OnkiHierarchyRelation> relations = new LinkedList<OnkiHierarchyRelation>();
	    	
	    	OnkiResource ores = this.createOnkiResource(uri, langPreferences);
	    	printRunTime("siblingTime", "1");
	    	//this.addConceptHierarchy(ores, langPreferences);
			if (parents != 0)
				this.addParentConcepts(ores, langPreferences);
			if (children != 0)
				this.addChildConcepts(ores, langPreferences);
			if (siblings != 0) {
				this.addSiblingConcepts(ores, langPreferences);
				printRunTime("siblingTime", "2");
			}
			if (parents != 0) {
				for (OnkiTreeResource relation : ores.getParentTree()) {
					String parentUri = null;
					if (relation.getOnkiResourceParent() != null)
						parentUri = relation.getOnkiResourceParent().getUri();
					// superclasses
					if (parents == -1)
						relations.add(new OnkiHierarchyRelation(relation.getOnkiResource().getUri(), parentUri, relation.getOnkiResource().getLabel(), relation.getTreeindentLevel()));
	
					if (parents == 1 && relation.getOnkiResource().equals(ores)) {
						OnkiHierarchyRelation rel = new OnkiHierarchyRelation(relation.getOnkiResource().getUri(), parentUri, relation.getOnkiResource().getLabel(), relation.getTreeindentLevel());
						if (!relations.contains(rel)) {
							if (parentUri != null)
								relations.add(new OnkiHierarchyRelation(parentUri, null, this.createOnkiResource(parentUri, langPreferences).getLabel(), relation.getTreeindentLevel()-1));
							relations.add(rel);
						}
					}
	
					/*if (relation.getOnkiResource().equals(ores)) {
						// sibling classes
						if (siblings == 1 && parentUri != null)
							for (OnkiResource sibling : ores.getSiblingsWithParent(relation.getOnkiResourceParent()))
								relations.add(new OnkiHierarchyRelation(sibling.getUri(), parentUri, sibling.getLabel(), relation.getTreeindentLevel()));
						// subclasses
						ores.getSubnodesWithIndent(relation.getTreeindentLevel(), children, relations);
					}*/
				}
			}
			
			// sibling classes
			if (siblings == 1) {
				printRunTime("siblingTime", "3");
				for (OnkiResource parent : ores.getParents()) {
					printRunTime("siblingTime", "22");
					for (OnkiResource sibling : ores.getSiblingsWithParent(parent)) {
						printRunTime("siblingTime", "24");
						relations.add(new OnkiHierarchyRelation(sibling.getUri(), parent.getUri(), sibling.getLabel(), -1));
						printRunTime("siblingTime", "23");
					}
				}
				printRunTime("siblingTime", "21");
			}
			// subclasses
			ores.getSubnodesWithIndent(-1, children, relations);
			
			return relations;
    	}
    	return null;
    }

    /**
	 * Is a concept a top concept of a SKOS ConceptScheme. 
	 */
    
    public boolean isTopConcept(Resource res) {
    	Property hasTopConcept = ResourceFactory.createProperty(SKOS_NS+"hasTopConcept");
    	if (ontologyModel.listStatements(null, hasTopConcept, res).hasNext())
    		return true;
    	return false;
    }

    /**
	 * Get the resources representing the SKOS ConceptSchemes a concept belongs to.
	 */
    
    public Collection<Resource> getConceptSchemes(Resource res) {
    	HashSet<Resource> resSet = new HashSet<Resource>();
    	Property hasTopConcept = ResourceFactory.createProperty(SKOS_NS+"hasTopConcept");
    	StmtIterator stmtIter = ontologyModel.listStatements(null, hasTopConcept, res);
    	while (stmtIter.hasNext())
    		resSet.add(stmtIter.nextStatement().getSubject());
    	Property inScheme = ResourceFactory.createProperty(SKOS_NS+"inScheme");
    	stmtIter = ontologyModel.listStatements(res, inScheme, (RDFNode)null);
    	while (stmtIter.hasNext())
    		resSet.add(stmtIter.nextStatement().getResource());
    	return resSet;
    }
    
    /**
	 * Get the resources representing the SKOS ConceptSchemes.
	 */
    
    public Collection<Resource> getConceptSchemes() {
    	HashSet<Resource> resSet = new HashSet<Resource>();
    	Resource conceptScheme = ResourceFactory.createResource(SKOS_NS+"ConceptScheme");
    	StmtIterator stmtIter = ontologyModel.listStatements(null, RDF.type, conceptScheme);
    	while (stmtIter.hasNext())
    		resSet.add(stmtIter.nextStatement().getSubject());
    	return resSet;
    }
    
    /**
	 * Get the top concepts of a SKOS ConceptScheme.
	 */
    
    public Collection<Resource> getTopConcepts() {
    	HashSet<Resource> resSet = new HashSet<Resource>();
    	for (Resource scheme : this.getConceptSchemes()) {
    		for (StmtIterator iter = scheme.listProperties(ResourceFactory.createProperty(SKOS_NS+"hasTopConcept")); iter.hasNext(); ) {
    			resSet.add((Resource)iter.nextStatement().getObject());
    		}
    	}
    	return resSet;
    }
    
    /**
	 * Does a concept have equivalent concepts. 
	 */
    
    public boolean hasEquivalentClass(Resource res) {
    	String[] equivClassProps = config.getStringArray(OnkiConfigurator.ELEM_CLASSES_EQUIVALENT_CLASS_PROPERTIES_PROPERTY);
    	for (int i=0; i<equivClassProps.length; i++) {
    		String equivClassProp =	equivClassProps[i];
    		if (OnkiConfigurator.isUsedInHierarchy(i)) {
	    		if (this.ontologyModel.listStatements(res, ResourceFactory.createProperty(equivClassProp), (RDFNode)null).hasNext())
	    			return true;
	    		// TODO: add configuration key for "symmetry of mapping relation"? (owl:equivalentClass: true)
	    		else if (this.ontologyModel.listStatements(null, ResourceFactory.createProperty(equivClassProp), res).hasNext())
	    			return true;
    		}
    	}
    	return false;
    }
    
    /**
	 * Get the equivalent concepts a concept has. 
	 */
    
    public Collection<Resource> getEquivalentClassesRes(Resource res, Collection<Resource> set) {
    	if (set == null)
    		set = new HashSet<Resource>();
    	else
    		set.add(res);
    	
    	if (this.hasEquivalentClass(res)) {
    		String[] equivClassProps = config.getStringArray(OnkiConfigurator.ELEM_CLASSES_EQUIVALENT_CLASS_PROPERTIES_PROPERTY);
    		for (int i=0; i<equivClassProps.length; i++) {
    			String equivClassProp =	equivClassProps[i];
    			if (OnkiConfigurator.isUsedInHierarchy(i)) {
		    		// case 1) direct statements
	    			for (StmtIterator iter = this.ontologyModel.listStatements(res, ResourceFactory.createProperty(equivClassProp), (RDFNode)null); iter.hasNext();) {
		    			RDFNode equivNode = iter.nextStatement().getObject();
		    			if (equivNode.isURIResource() && !set.contains((Resource)equivNode)) {
		    				// case 1b) transitivity 
		    				if (true) // TODO: add configuration key for "transitivity mapping relation" (owl:equivalentClass: true)
		    					this.getEquivalentClassesRes((Resource)equivNode, set);
		    				else
		    					set.add((Resource)equivNode);
		    			}
		    		}
		    		
		    		// case 2) inversed statements (property is symmetric)
		    		// TODO: add configuration key for "symmetry of mapping relation"? (owl:equivalentClass: true)
	    			for (StmtIterator iter2 = this.ontologyModel.listStatements(null, ResourceFactory.createProperty(equivClassProp), res); iter2.hasNext();) {
		    			RDFNode equivNode = iter2.nextStatement().getSubject();
		    			if (equivNode.isURIResource() && !set.contains((Resource)equivNode)) {
		    				//case 2b) transitivity
		    				if (true) // TODO: add configuration key for "transitivity of mapping relation" (owl:equivalentClass: true)
		    					this.getEquivalentClassesRes((Resource)equivNode, set);
		    				else
		    					set.add((Resource)equivNode);
		    			}
		    		}
    			}
    		}
    	}
    	return set;
    }
    
    /**
	 * Get the directory view of the ontology.
	 */
    
    public OnkiDirectory getDirectory(String type, List<String> langPreferences, int maxHits) {
    	OnkiDirectory directory = new OnkiDirectory();
    	
    	String lang = langPreferences.get(0);
    	if (lang == null)
    		lang = "";
    	
    	// default directory type priority: 1), 2), 3)  
    	if (type == null && config.containsKey(OnkiConfigurator.ELEM_CLASSES_GROUP_ENABLED) && config.getBoolean(OnkiConfigurator.ELEM_CLASSES_GROUP_ENABLED))
    		type = GROUPS;
    	
    	// 1) group list (case YSA)
    	if (type != null && type.equals(GROUPS)) {
    		
    		if (directories.containsKey(GROUPS) && directories.get(GROUPS).containsKey(lang) && maxHits == DEFAULT_DIRECTORY_SIZE)
    			return directories.get(GROUPS).get(lang);
    		
    		for (IndexResult result : this.searchConcept("*", lang, null, null, null, true, false, -1, -1).getResults()) {
	    		String groupUri = result.getUri();
	    		String groupLabel = this.createOnkiResource(groupUri, langPreferences).getLabel();
	    		
	    		LinkedList<OnkiResourceSerializable> members = new LinkedList<OnkiResourceSerializable>();
	    		LinkedList<String> groups = new LinkedList<String>();
	    		groups.add(groupUri);
	    		// get all the hits to get the total hit amount and to base the sorting to the prefLabels 
	    		IndexResults results = this.searchConcept("*", lang, null, null, groups, false, false, -1, -1);
	    		if (maxHits != 0) {		    		
		    		for (IndexResult member : results.getResults()) {
		    			members.add(new OnkiResourceSerializable(member.getUri(), member.getPrefLabel()));
		    			if (members.size() >= maxHits)
		    				break;
		    		}
	    		}
	    		directory.addElement(new OnkiDirectoryElement(groupUri, groupLabel, results.getMetadata().getTotalHitsAmount(), members));
	    	}
	    }
    	else {
    		// 2) top concepts (skos)	
    		if (type == null || type.equals(TOPCONCEPTS)) {

    			if (directories.containsKey(TOPCONCEPTS) && directories.get(TOPCONCEPTS).containsKey(lang) && maxHits == DEFAULT_DIRECTORY_SIZE)
    				return directories.get(TOPCONCEPTS).get(lang);
    			
	    		Collection<Resource> topConcepts = this.getTopConcepts();
	    		if (topConcepts.size() > 0 && topConcepts.size() <= 50) {
	    			type = TOPCONCEPTS;
	    			for (Resource topConcept : topConcepts) {
	    				String topConceptUri = topConcept.getURI();
	    				LinkedList<String> tempLangPrefs = new LinkedList<String>();
	    				tempLangPrefs.add(lang);
	    				String topConceptLabel = this.createOnkiResource(topConceptUri, tempLangPrefs).getLabel();
	    				if (!topConceptLabel.equals(topConcept.getLocalName()+"*")) {
		    				LinkedList<OnkiResourceSerializable> members = new LinkedList<OnkiResourceSerializable>();
		    				LinkedList<String> parents = new LinkedList<String>();
		    				parents.add(topConcept.getURI());
		    				IndexResults results = this.searchConcept("*", lang, null, parents, null, false, false, -1, -1);
		    				if (maxHits != 0) {		    		
		    		    		for (IndexResult member : results.getResults()) {
		    		    			members.add(new OnkiResourceSerializable(member.getUri(), member.getPrefLabel()));
		    		    			if (members.size() >= maxHits)
		    		    				break;
		    		    		}
		    	    		}
		    				directory.addElement(new OnkiDirectoryElement(topConceptUri, topConceptLabel, results.getMetadata().getTotalHitsAmount(), members));
	    				}
	    			}
	    		}
    		}
    		// 3) alphabetical index
    		if ((type != null && type.equals(ALPHABETICAL)) || type == null) {
    			type = ALPHABETICAL;
    			
	    		if (directories.containsKey(ALPHABETICAL) && directories.get(ALPHABETICAL).containsKey(lang) && maxHits == DEFAULT_DIRECTORY_SIZE)
	    			return directories.get(ALPHABETICAL).get(lang);
	    		
	    		for (char alphabet : this.getAvailableAlphabets(lang)) {
	    			LinkedList<OnkiResourceSerializable> members = new LinkedList<OnkiResourceSerializable>();
	    			// get all the hits to get the total hit amount and to base the sorting to the prefLabels
					IndexResults results = this.searchConcept(String.valueOf(alphabet), lang, null, null, null, false, false, -1, -1);
	    			if (maxHits != 0) {
			    		for (IndexResult member : results.getResults()) {
			    			// TODO: prefLabel should be normalized before comparison (i.e. now concepts starting with diacritics are not added to directory)
			    			if (member.getPrefLabel().length() > 0 && member.getPrefLabel().toLowerCase().charAt(0) == alphabet)
			    				members.add(new OnkiResourceSerializable(member.getUri(), member.getPrefLabel()));
			    			if (members.size() >= maxHits)
			    				break;
			    		}
	    			}
	    			// FIXME: totalHits is not correct value if there is matches to altLabels 
	    			if (maxHits == 0 || members.size() > 0)
	    				directory.addElement(new OnkiDirectoryElement(null, String.valueOf(alphabet), results.getMetadata().getTotalHitsAmount(), members));
	    		}
		    }
    	}
    	// TODO: 4) representative concepts of ontology (case YSO)
    	
    	// if the directory of desired lang is empty, use lang priority list 
    	if (directory.isEmpty() && langPreferences.size() > 1) {
    		langPreferences.remove(0);
    		return this.getDirectory(type, langPreferences, maxHits);
    	}
    	
    	directory.setType(type);
    	
    	if (!directory.isEmpty()) {
    		directory.sort();
    		directory.setLang(lang);
	    	if (maxHits == DEFAULT_DIRECTORY_SIZE && (lang.equals("fi") || lang.equals("en") || lang.equals("sv"))) {
	    		if (!directories.containsKey(directory.getType()))
	    			directories.put(directory.getType(), new HashMap<String, OnkiDirectory>());
	    		directories.get(directory.getType()).put(lang, directory);
	    	}
	    }

    	return directory;
    }
    
    /**
	 * Generate default directory views for the ontology (cached for ONKI2 use).
	 */

    public void generateDefaultDirectoryViews() {
    	logger.info("generateDefaultViews() Generating default ontology directory views");
    	long startTime = System.currentTimeMillis();

    	File serFile = null;
    	
    	if (config.getString(OnkiConfigurator.ELEM_INDEX_TYPE).equals("file"))
    	   	serFile = new File(config.getString(OnkiConfigurator.ELEM_INDEX_PATH), "directory-views.ser");
    	
    	boolean newSerialization = true;
    	
    	if (serFile != null && serFile.exists() && serFile.isFile() && serFile.lastModified() > this.filesLastModified) { // serialized object is available and timestamp newer than rdf file's
    		// restore directory views from disk serialization
    		try {
    			ObjectInputStream in = new ObjectInputStream(new FileInputStream(serFile));
    			this.directories = (HashMap<String, HashMap<String, OnkiDirectory>>)in.readObject();
    			in.close();
    			newSerialization = false;	
    		}
    		catch (IOException e) {
    			logger.error("generateDefaultDirectoryViews() : error reading serialization file.");
    		}
    		catch (ClassNotFoundException e) {
    			logger.info("generateDefaultDirectoryViews() : class version mismatch when restoring serialized object.");
    		}
    	}

    	if (newSerialization) {
	    	// load default ontology directory views to memory
		    LinkedList<String> langs = new LinkedList<String>();
		    langs.add("fi");
		    langs.add("sv");
		    langs.add("en");
		    
		    for (String lang : langs) {
		    	List<String> langList = this.getLangPreferences(lang);
		    	this.getDirectory(GROUPS, langList, DEFAULT_DIRECTORY_SIZE);
		    	this.getDirectory(TOPCONCEPTS, langList, DEFAULT_DIRECTORY_SIZE);
		    	this.getDirectory(ALPHABETICAL, langList, DEFAULT_DIRECTORY_SIZE);
		    }

		    // TODO: should the "directories" object manipulation be done here?

		    // serialize directory views to disk
		    if (serFile != null) {
			    try {
			    	serFile.delete();
			    	ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(serFile));
			    	out.writeObject(this.directories);
			    	out.close();
			    }
			    catch (IOException e) {
			    	logger.error("generateDefaultDirectoryViews() : error writing serialization file.");
			    }
		    }
    	}
	    long endTime = System.currentTimeMillis();
        logger.info("generateDefaultViews() Default ontology directory views generated in in " + ((double) (endTime - startTime)) / 1000.0 + " secs.");
    }
    
    /**
	 * Get a resource as RDF output.
	 */
    
    public void getResourceAsRDF(String uri, OutputStream out) {
    	Model model = ModelFactory.createDefaultModel();
    	model.setNsPrefixes(this.ontologyModel.getNsPrefixMap());
    	OnkiResource ores = this.createOnkiResource(uri, null);
    	
    	// 1) get the statements from the current OnkiResource
    	for (OnkiStatement stmt : ores.getProperties()) {
    		Resource subj = stmt.getSubject().getResource();
    		Property pred = model.createProperty(stmt.getPredicate().getUri());
    		RDFNode obj = stmt.getObject();
    		model.add(model.createStatement(subj, pred, obj));
    	}
    	
    	// 2) get the statements where the current resource is object
    	for (StmtIterator iter = ontologyModel.listStatements(null, null, ores.getResource()); iter.hasNext(); )
    		model.add(iter.nextStatement());
    	
    	model.write(out);
    }
    
    // ------------ INDEXING BEGINS HERE -------------------------------------------------    
    
    /**
     * Index ontologies (with Lucene).
     */
    
    public void indexOntologies() {
    	long startTime = System.currentTimeMillis();
    	
    	String defaultSortingLang = this.config.getString(OnkiConfigurator.ELEM_LANGUAGES_PREFERENCE);
    	
    	if (config.getString(OnkiConfigurator.ELEM_INDEX_TYPE).equals("file"))
    		this.ontologyIndex = new AlphabeticalIndex(config.getString(OnkiConfigurator.ELEM_INDEX_PATH), this.filesLastModified, defaultSortingLang);
    	else
    		this.ontologyIndex = new AlphabeticalIndex(defaultSortingLang);
    	
    	Set<String> searchLabels = new HashSet<String>();
    	searchLabels.addAll(config.getList(OnkiConfigurator.ELEM_CLASSES_INDEX_LABEL));
    	if (searchLabels.isEmpty()) // index (UI) labels only if index labels are not specified
    		searchLabels.addAll(config.getList(OnkiConfigurator.ELEM_CLASSES_LABEL));
    	
    	Set<Property> properties = new HashSet<Property>();
    	for (String searchLabel : searchLabels) {
    		logger.info("Searchlabel: " + searchLabel + "\n");
    		try {
    			properties.add((Property)this.ontologyModel.getProperty(searchLabel));
    		} catch(Exception ex) {
    			System.out.println("Catched exception "+ex+" when resolving properties during indexing of ontologies.");
    		}
    	}

    	if (properties.size() == 0)
    		return;
    	
    	String[] searchAltLabels = config.getStringArray(OnkiConfigurator.ELEM_CLASSES_INDEX_ALT_LABEL);
    	Set<Property> altProperties = new HashSet<Property>();
    	for(int i=0; i<searchAltLabels.length; i++) {
    		logger.info("SearchAltlabel: " + searchAltLabels[i] + "\n");
    		try {
    			altProperties.add((Property)this.ontologyModel.getProperty(searchAltLabels[i]));
    		} catch(Exception ex) {
    			System.out.println("Catched exception "+ex+" when resolving altProperties during indexing of ontologies.");
    		}
    	}
    	
		String[] searchHiddenLabels = config.getStringArray(OnkiConfigurator.ELEM_CLASSES_INDEX_HIDDEN_LABEL);
		Set<Property> hiddenProperties = new HashSet<Property>();
		for(int i=0; i<searchHiddenLabels.length; i++) {
			logger.info("SearchHiddenlabel: " + searchHiddenLabels[i] + "\n");
			try {
				hiddenProperties.add((Property)this.ontologyModel.getProperty(searchHiddenLabels[i]));
			} catch(Exception ex) {
				System.out.println("Catched exception "+ex+" when resolving hiddenProperties during indexing of ontologies.");
			}
		}
    	
    	Collection<String> typeUris = new LinkedList<String>();
    	if (config.containsKey(OnkiConfigurator.ELEM_CLASSES_TYPE_RESTRICTION_TYPE))
    		typeUris = Arrays.asList(config.getStringArray(OnkiConfigurator.ELEM_CLASSES_TYPE_RESTRICTION_TYPE));

    	boolean typeUrisInclusive = false;
    	if (config.containsKey(OnkiConfigurator.ELEM_CLASSES_TYPE_RESTRICTION_INCLUSIVE))
    		typeUrisInclusive = config.getBoolean(OnkiConfigurator.ELEM_CLASSES_TYPE_RESTRICTION_INCLUSIVE);
    	
    	IndexItems ii = new IndexItems(this.ontologyModel.listStatements(), properties, altProperties, hiddenProperties, typeUris, typeUrisInclusive);
    	String[] namespaces = config.getStringArray(OnkiConfigurator.ELEM_CLASSES_INDEX_NAMESPACE);
    	Set<String> nsSet = new HashSet<String>(Arrays.asList(namespaces));
		
    	this.ontologyIndex.initialize(ii, nsSet);
    	
    	System.out.println("Ontologyindex created: "+this.ontologyIndex);
    	long endTime = System.currentTimeMillis();
        logger.info("indexOntologies() Ontologies indexed in " + ((double) (endTime - startTime)) / 1000.0 + " secs.");
    }
    
    /**
     * A class for representing IndexItems as an iterable collection.
     * 
     * Please note: The current implementation does allow subjects to be indexed several times, 
     * for each matching property! This might be a feature or bug depending on the situation.
     * 
     * @author kimvilja
     */
    
    private class IndexItems implements Iterable<IndexItem> {

    	private Iterator<IndexItem> iterator;
    	
    	private IndexItems(StmtIterator stmti, Collection<Property> properties, Set<Property> altProperties, Set<Property> hiddenProperties, Collection<String> typeUris, boolean typeUrisInclusive) {
    		this.iterator = new IndexItemsIterator(stmti,properties,altProperties,hiddenProperties,typeUris,typeUrisInclusive);
    	}

		public Iterator<IndexItem> iterator() {
			return this.iterator;
		}
		
		private class IndexItemsIterator implements Iterator<IndexItem> {
			
			private StmtIterator stmtiterator;
			private Collection<Property> properties;
			private Iterator<Property> propertyiterator;
			
			private Set<Property> altProperties;
			private Set<Property> hiddenProperties;

			private Statement currentstatement = null;
			private IndexItem nextindexitem=null;
			private boolean started=false;
			private boolean closed=false;
			
			private Collection<String> typeUris;
			private boolean typeUrisInclusive;
			
			private IndexItemsIterator(StmtIterator stmtiterator, Collection<Property> properties, Set<Property> altProperties, Set<Property> hiddenProperties, Collection<String> typeUris, boolean typeUrisInclusive) {
				this.stmtiterator = stmtiterator;
				this.currentstatement = this.stmtiterator.nextStatement();
				this.properties = properties;
				this.propertyiterator = properties.iterator();
				this.altProperties = altProperties;
				this.hiddenProperties = hiddenProperties;				
				this.typeUris = typeUris;
				this.typeUrisInclusive = typeUrisInclusive;
			}

			public boolean hasNext() {
				if (this.closed)
					return false;
				if (!this.started)
					this.next();
				return this.nextindexitem != null;
			}
			
			/**
			 * Return the next IndexItem.
			 */
			
			public IndexItem next() {
				IndexItem next = null;
				if (this.started && this.nextindexitem == null || this.closed)
					return null;

				if (this.started)
					next = this.nextindexitem;

				this.started = true;
				this.nextindexitem = null;
				
				// subjectiterator
				while (this.nextindexitem==null && 
						this.currentstatement != null && 
						this.stmtiterator.hasNext()) {

					boolean notIndexed = false;

					// propertyiterator
					while(this.currentstatement.getSubject().isURIResource() &&
							this.nextindexitem== null && 
							this.currentstatement != null && 
							this.propertyiterator.hasNext()) {
						Property p = this.propertyiterator.next();
						
						if (this.currentstatement.getPredicate().equals(p)) {
							RDFNode object = this.currentstatement.getObject();
							if (object.isLiteral()) {
								Literal lit = (Literal)object;
								OnkiResource ores = new OnkiResource(this.currentstatement.getSubject(), null);
								
								// Get all alternative labels for this item
								List<String> altLabels = getAltLabels(ores, lit, false);
								List<String> hiddenLabels = getAltLabels(ores, lit, true);
								
								// TODO: OnkiResource could know how to create an IndexItem of a given OnkiResource?
								// get types
								ArrayList<String> types = new ArrayList<String>();
								boolean isGroup = false;
								
								// support for groups
								if (ores.isGroup()) {						
									isGroup = true;
								}
								else {
									Iterator<OnkiResource> typeIter = createOnkiResource(this.currentstatement.getSubject().getURI(), null).getTypes().iterator();
									
									// if a concept is not typed and there typeUris to include are defined, the concept is no indexed 
									if (!typeIter.hasNext() && this.typeUrisInclusive && this.typeUris.size() > 0)
										notIndexed = true;
									
									while (typeIter.hasNext()) {
										String type = typeIter.next().getUri();
										
										if (type != null) { // don't use blank nodes

											// if one of the types of concept is excluded, the concept is not indexed at all 
											if ((this.typeUrisInclusive && !this.typeUris.contains(type)) || (!this.typeUrisInclusive && this.typeUris.contains(type))) {
												notIndexed = true;
												break;
											}
											types.add(type);
										}
									}
								}

								// get super classes
								// TODO: set data structure would be better to remove the duplicates?
								ArrayList<String> superclasses = new ArrayList<String>();
								Iterator<OnkiResource> superClassIter = createOnkiResource(this.currentstatement.getSubject().getURI(), null).getAllParents().iterator();
								while (superClassIter.hasNext()) {
									String uri = superClassIter.next().getUri();
									if (!this.currentstatement.getSubject().getURI().equals(uri))
										superclasses.add(uri);
								}

								// get "groups"
								ArrayList<String> groups = new ArrayList<String>();
								Iterator<OnkiResource> groupIter = createOnkiResource(this.currentstatement.getSubject().getURI(), null).getGroups().iterator(); 
								while (groupIter.hasNext()) 
									groups.add((groupIter.next()).getUri());
								
								// get "query expansions"
								ArrayList<String> queryExpansions = new ArrayList<String>();
								Iterator<OnkiResource> queryExpansionIter = createOnkiResource(this.currentstatement.getSubject().getURI(), null).getQueryExpansions().iterator();
								while (queryExpansionIter.hasNext()) {
									String qeUri = queryExpansionIter.next().getUri();
									if (qeUri != null) // don't use blank nodes
										queryExpansions.add(qeUri);
								}
								
								this.nextindexitem = new IndexItem(this.currentstatement.getSubject().getNameSpace(),
										ores.getNamespacePrefix(),
										this.currentstatement.getSubject().getLocalName(),
										lit.getLanguage(),
										lit.getString(),
										altLabels,
										hiddenLabels,
										types,
										superclasses,
										groups,
										isGroup,
										queryExpansions);
							}
							break;
						}
					}

					if (notIndexed)
						this.nextindexitem = null;
					
					if (this.nextindexitem==null && this.stmtiterator.hasNext()) {
						this.currentstatement = this.stmtiterator.nextStatement();
						this.propertyiterator = this.properties.iterator(); // back to beginning
					}
					// does not work?
					if (notIndexed)
						continue;
				}

				if (this.nextindexitem == null)
					this.closed = true;

				return next;
			}
			
			public void remove() {
				// TODO Auto-generated method stub
				
			}
			
			private List<String> getAltLabels(OnkiResource ores, Literal prefLabelLit, boolean hiddenLabels) {
				ArrayList<String> altLabels = new ArrayList<String>();
				Set<Property> altLabelProperties;
				if (hiddenLabels)
					altLabelProperties = this.hiddenProperties;
				else
					altLabelProperties = this.altProperties;
				
				for (OnkiStatement stmt: ores.getProperties(false, altLabelProperties, false, false)) {
					RDFNode altLabelObject = stmt.getObject();
					if (altLabelObject.isLiteral()) {
						Literal altLabelLit = (Literal)altLabelObject;
						// should the lang-codeless altLabels be indexed as alternative to all prefLabels?
						if (altLabelLit.getLanguage().equals(prefLabelLit.getLanguage()) &&
								!stmt.getObjectAsString().equals(prefLabelLit.getString())) // don't index altLabel if altLabel == prefLabel
							altLabels.add(stmt.getObjectAsString());
					}
					// get alternative labels from container structures (e.g. in TGN)
					// TODO: maybe should be more general Container instead of Seq
					else {
						try {
							//if (altLabelObject.canAs(java.lang.Class.forName("com.hp.hpl.jena.rdf.model.Seq"))) {
							if (altLabelObject.canAs((Class<Seq>)java.lang.Class.forName("com.hp.hpl.jena.rdf.model.Seq"))) {
								//NodeIterator iter = ((Seq)altLabelObject.as(java.lang.Class.forName("com.hp.hpl.jena.rdf.model.Seq"))).iterator();						
								NodeIterator iter = ((Seq)altLabelObject.as((Class<Seq>)java.lang.Class.forName("com.hp.hpl.jena.rdf.model.Seq"))).iterator();
								while (iter.hasNext()) {
									RDFNode node = iter.nextNode();
									if (node.isLiteral()) {
										Literal altLabelLit = (Literal)node;
										if (altLabelLit.getLanguage().equals(prefLabelLit.getLanguage()) &&
											!altLabelLit.getString().equals(prefLabelLit.getString())) // don't index altLabel if altLabel == prefLabel
											altLabels.add(altLabelLit.getString());
									}
								}
							}
						} catch (ClassNotFoundException e) {
							System.err.println("Exception catched in IndexItem.next() when instantiating class object of com.hp.hpl.jena.rdf.model.Seq: "+e);
						}
					}
				}
				return altLabels;
			}
		}
    }

    //  ------------ INDEXING ENDS HERE -------------------------------------------------
    
    /**
	 * Search the concept index.
	 */

    public IndexResults searchConcept(String query, String lang, Collection<String> types, Collection<String> parents, Collection<String> groups, boolean isGroup, boolean isRoot, int maxHits, int startFromHit) {
    	return searchConcept(query, lang, types, parents, groups, isGroup, isRoot, maxHits, startFromHit, null, null, null, false);
    }
    
    public IndexResults searchConcept(String query, String lang, Collection<String> types, Collection<String> parents, Collection<String> groups, boolean isGroup, boolean isRoot, int maxHits, int startFromHit, Collection<String> excludedTypes, Collection<String> excludedParents, Collection<String> excludedGroups, boolean multipleHitsOfConcept) {
    	if (!config.containsKey(OnkiConfigurator.ELEM_CLASSES_INDEX_NAMESPACES_LANG) || !config.getBoolean(OnkiConfigurator.ELEM_CLASSES_INDEX_NAMESPACES_LANG))
    		lang = null;
    	if (this.ontologyIndex != null)
    		return this.ontologyIndex.search(lang, query, types, parents, groups, isGroup, isRoot, maxHits, startFromHit, excludedTypes, excludedParents, excludedGroups, multipleHitsOfConcept);
    	else {
    		logger.error("searchConcept() : index not ready");
    		return null;
    	}
    }
    
    /**
	 * Expand the query.
	 */
    
    public IndexResults expandQuery(Collection<String> uris, String lang, int maxHits, Collection<String> types) {
    	if (!config.containsKey(OnkiConfigurator.ELEM_CLASSES_INDEX_NAMESPACES_LANG) || !config.getBoolean(OnkiConfigurator.ELEM_CLASSES_INDEX_NAMESPACES_LANG))
    		lang = null;
    	if (this.ontologyIndex != null)
    		return this.ontologyIndex.expandQuery(uris, lang, maxHits, types);
    	else {
    		logger.error("searchConcept() : index not ready");
    		return null;
    	}
    }    
    
    // TEMPORARY PLACE FOR DEMO FOR CO-OCCURENCE MAPPING GENERATION UI, Jouni 15.9.2008
    public Collection<ConceptMapping> getYsoCoOccurences(List langPreferences) {
    	String SPARQLPrefixes = "PREFIX setmapping:  <http://www.yso.fi/onto/mapping/>"
			+ "PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>";
    	
    	String sparql =

    		SPARQLPrefixes
    				+ "SELECT DISTINCT ?argument1 ?argument2 ?covers ?coveredby "
    				+ " WHERE { {?mapinstance "
    				+ "              setmapping:argument1 ?argument1 ;"
    				+ "              setmapping:argument2 ?argument2 ;"
    				+ "              setmapping:covers ?covers ;"
    				+ "              setmapping:coveredBy ?coveredby ."
    				
    				+
    				// " FILTER ((xsd:decimal(?covers) * xsd:decimal(?coveredby)) > (xsd:decimal(0.05)))"
    				"}" + "} " + 
    				"ORDER BY DESC(xsd:decimal(?covers) * xsd:decimal(?coveredby))";
    	
    	com.hp.hpl.jena.query.Query query = QueryFactory.create(sparql);
    	com.hp.hpl.jena.query.QueryExecution qe = QueryExecutionFactory.create(query, this.ontologyModel);
		ResultSet results = qe.execSelect();
    	
		LinkedList<ConceptMapping> mappings = new LinkedList<ConceptMapping>();
		
		while (results.hasNext()) {
			QuerySolution qs = results.nextSolution();
			RDFNode argNode1 = qs.get("argument1");
			RDFNode argNode2 = qs.get("argument2");
			RDFNode coversNode = qs.get("covers");
			RDFNode coveredByNode = qs.get("coveredby");
			if (argNode1.isURIResource() && argNode2.isURIResource() && coversNode.isLiteral() && coveredByNode.isLiteral()) {
				Resource arg1 = (Resource)argNode1;
				Resource arg2 = (Resource)argNode2;
				double covers = ((Literal)coversNode).getDouble();
				double coveredBy = ((Literal)coveredByNode).getDouble();
				
				// temporary hack
				boolean redundant = false;
				Property related = ResourceFactory.createProperty("http://www.yso.fi/onto/yso-meta/2007-03-02/associativeRelation");
				
				if (ontologyModel.listStatements(arg1, RDFS.subClassOf, arg2).hasNext())
					redundant = true;
				else if (ontologyModel.listStatements(arg2, RDFS.subClassOf, arg1).hasNext())
					redundant = true;
				else if (ontologyModel.listStatements(arg1, related, arg2).hasNext())
					redundant = true;
				else if (ontologyModel.listStatements(arg2, related, arg1).hasNext())
					redundant = true;
				else {
					for (StmtIterator iter = ontologyModel.listStatements(arg1, RDFS.subClassOf, (RDFNode)null); iter.hasNext(); ) {
						if (ontologyModel.listStatements(arg2, RDFS.subClassOf, iter.nextStatement().getResource()).hasNext())
							redundant = true;
					}
				}
				if (!redundant)
					mappings.add(new ConceptMapping(new OnkiResource(arg1,  langPreferences), new OnkiResource(arg2, langPreferences), covers, coveredBy));
			}
		}
		return mappings;
    }
    
    public class ConceptMapping {
    	private OnkiResource arg1;
    	private OnkiResource arg2;
		private double overlaps;
		private double overlappedBy;
		
		public ConceptMapping(OnkiResource arg1, OnkiResource arg2, double overlaps, double overlappedBy) {
			this.arg1 = arg1;
			this.arg2 = arg2;
			this.overlaps = overlaps;
			this.overlappedBy = overlappedBy;
		}
		
		public OnkiResource getArg1() {
			return this.arg1;
		}
		
		public OnkiResource getArg2() {
			return this.arg2;
		}
		
		public double getOverlaps() {
			return this.overlaps;
		}
		
		public double getOverlappedBy() {
			return this.overlappedBy;
		}
		
		public BigDecimal getOverlapsRounded() {
			return new BigDecimal(this.overlaps).round(new MathContext(3));
		}
		
		public BigDecimal getOverlappedByRounded() {
			return new BigDecimal(this.overlappedBy).round(new MathContext(3));
		}
		
		public double getConfidence() {
			return this.overlaps * this.overlappedBy;
		}
		
		public BigDecimal getConfidenceRounded() {
			return new BigDecimal(this.overlaps * this.overlappedBy).round(new MathContext(3));
		}
    }
    
    
    public String getConceptLabel(String uri, List langPreferences) {
    	if (uri == null)
    		return null;
    	
		String lang = null;
		if (!langPreferences.isEmpty())
			lang = (String)langPreferences.get(0);
		String key = generateuriToLabelCacheKey(uri, lang);

		// try to get label from cache
		Object labelObj = this.uriToLabelCache.get(key);
		if (labelObj != null) {
			cacheHits++;
			return (String)labelObj;
		}

		cacheMisses++;

		// try to get label from Lucene if the indexed labels are meant for UI
		// presumption: if the index contains a concept label in some language it contains all the available labels of the concept
		if (this.ontologyIndex != null && this.getConfig().getList(OnkiConfigurator.ELEM_CLASSES_INDEX_LABEL).isEmpty()) {
			this.printRunTime("siblingTime", "65");

    		Map<String, String> labels = this.ontologyIndex.getLabels(uri);
    		if (labels != null) {
	    		for (Iterator langI = langPreferences.iterator(); langI.hasNext(); ) {	    			
	    			String currentLang = (String)langI.next();
	    			String label = null;
	    			if (currentLang == null) {
	    				label = labels.values().iterator().next();
	    			}
	    			else if (labels.containsKey(currentLang)) {
	    				label = labels.get(currentLang);
	    			}
	    			if (label != null) {
	    				String currentKey = generateuriToLabelCacheKey(uri, currentLang);
	    				this.uriToLabelCache.add(currentKey, label);
	    				if (currentLang != null && !currentLang.equals(lang))
		    				this.uriToLabelCache.add(key, label);
	    				this.printRunTime("siblingTime", "66");
		    			return label;
	    			}
	    		}
    		}
		}
		
		// try to get label from Jena
		for (Iterator langI = langPreferences.iterator(); langI.hasNext(); ) {
        	String currentLang = (String)langI.next();
		    for (Iterator iter = this.getLabelProps().iterator();  iter.hasNext();) {
		        Property p = (Property) iter.next();
		        try {
		        	if (this.getConcept(uri) != null) {
//    		        		processor.printRunTime("siblingTime", "46");
				        for (StmtIterator stmtIter = this.getConcept(uri).listProperties(p); stmtIter.hasNext();) {
//    				        	processor.printRunTime("siblingTime", "47");
				            Statement stmt = stmtIter.nextStatement();
//    				            processor.printRunTime("siblingTime", "50");
				            if (stmt == null) continue;
		
				            if (stmt.getLanguage() != null && stmt.getLanguage().equals(currentLang) ||
				            	currentLang == null) {
				            	this.printRunTime("siblingTime", "51");
				            	
				            	String label = stmt.getString();
				            	String currentKey = generateuriToLabelCacheKey(uri, currentLang);
				    			this.uriToLabelCache.add(currentKey, label);
				    			if (currentLang != null && !currentLang.equals(lang)) {
				    				this.uriToLabelCache.add(key, label);
				    			}

				            	return label;
				        	}
//    				            processor.printRunTime("siblingTime", "48");
				        }
		        	}
		        } catch(HasNoModelException e) {} // Occurs when the property is not connected with a model.
		    }
        }
        this.printRunTime("siblingTime", "54");

        if (this.getConcept(uri) != null) {
        	String label = this.getConcept(uri).getLocalName()+"*";
			this.uriToLabelCache.add(key, label);
        	return label;
        }
    	return null;
    }
    
    public void printUriToLabelCacheStats() {
    	this.uriToLabelCache.printStats();
    }
    
	private String generateuriToLabelCacheKey(String uri, String lang) {
		String key;
		if (lang != null)
			key = uri+"\""+lang;
		else
			key = uri;
		return key;
	}
    
    /**
     * A class for caching resources, e.g. labels for concepts.
     * 
     */
	
    private class Cache {

    	private String cacheRegionName;
    	private JCS cache = null;
    	private Logger logger;
    	
    	private Cache(String cacheRegionName) {
    		try {
    			this.cacheRegionName = cacheRegionName;
    			this.cache = JCS.getInstance(this.cacheRegionName);
    	        this.logger = Logger.getLogger(getClass());
    		}
    		catch (CacheException e) {
    			this.logger.error("Cache() : Problem initializing cache for region name [" + this.cacheRegionName + "].");
    		}
    	}
    	   	
    	private Object get(String key) {
    		return this.cache.get(key);
    	}
    	
    	private void add(String key, Object obj) {
    		if (key != null && obj != null) {
				try {
					this.cache.put(key, obj);
				}
				catch(CacheException e) {
					this.logger.error("add() : Problem putting " + key + " in the cache.");
				}
    		}
    	}
    	
    	private void remove(String key) {
    		if (key != null) {
    			try {
    				this.cache.remove(key);
    			}
    			catch(CacheException e) {
    				this.logger.error("remove() : Problem removing " + key + " from the cache");
    			}
    		}
    	}
    	
    	private boolean contains(String key) {
    		return (this.cache.get(key) != null);
    	}
    	
    	private void printStats() {
        	IStats[] cacheStatsArray = this.cache.getStatistics().getAuxiliaryCacheStats();
        	for (int i=0; i<cacheStatsArray.length; i++) {
        		IStatElement[] statElementArray = cacheStatsArray[i].getStatElements();  
    	    	for (int j=0; j<statElementArray.length; ++j) {
    	    		System.out.println(statElementArray[j].getName()+" - "+statElementArray[j].getData());
    	    	}
        	}
    	}
    	
    	private void clear() {
        	try {
        		this.cache.clear();
        	}
        	catch (CacheException e) {
        		this.logger.error("clear() : Problem clearing cache");
        	}
        }
    }
    
    // Debug printing and logging facilities
    
    public static void logRunTime(boolean printToFile) {
    	runTimePrintLog = printToFile;
    	if (!printToFile)
    		try {
    			runTimeLogStream.close();
    			runTimeLogStream = null;
    		} catch(Exception e) {
    			System.err.println("log file could not be closed");
    		}
    }
    
    public static void printRunTime(String key, String id) {
    	printRunTime(key, id, false, false);
    }
    
    public static void printRunTime(String key, String id, boolean block) {
    	printRunTime(key, id, block, false);
    }
    
    public static void printRunTime(String key, String id,  boolean block, boolean reset) {
    	printRunTime(key, id, -1, block, reset);
    }
    
    public static void printRunTime(String key, String id, long time, boolean block, boolean reset) {
    	if (debugLogEnabled) {
	    	if (reset) {
	    		startRunTime = System.nanoTime();
	    		prevRunTime = startRunTime;
	    		prevRunTimeId = "";
	    		return;
	    	}
	    	
	    	if (runTimeLogStream == null) {
	    		try {
	    			runTimeLogStream = new FileOutputStream("/home/jwtuomin/tmp/onki-runtime.log", true);
	    		} catch(Exception e) {
	    			System.err.println("log file could not be created");
	    		}
	    	}
	    	
	    	long currentTime = System.nanoTime();
	    	
	    	String prefix = "";
	    	if (block) {
		    	prefix = "B";
		    	if (time != -1)
					prefix = "O";
	    	}
	    	else
	    		prefix = "T";
	    	
	    	if (time == -1)
				time = (long)((currentTime-startRunTime) / 1000000.0);
	    	
	    	double delta = (currentTime-prevRunTime) / 1000000.0;
	    	if (runTimePrintLog || block) {
	    		try {
	    			String output = null;
	    			if (block)
	    				output = (prefix+"\t"+key+"\t"+id+"\t\t"+ time +"\n");
	    			else if (!runTimePrintLogOnlyBlocks)
	    				output = (prefix+"\t"+key+"\t"+id+"\t"+prevRunTimeId+"\t"+delta+"\n");
	    			if (block || !runTimePrintLogOnlyBlocks)
	    				runTimeLogStream.write(output.getBytes());
	    		}
	    		catch(Exception e) {
	    			System.err.println("could not print to log file");
	    		}
	    	}
	    	if (debugToStdOut) {
		    	if (block)
		    		System.out.println(key+" "+id+":\t"+ time);
		    	else if (delta >= 1) {
		    		System.out.print(key+" "+id+" ("+prevRunTimeId+"):\t"+ delta +", "+ (currentTime-startRunTime) / 1000000.0+" \t");
			    	int starAmount = (int) delta / 10;
			    	for (int i = 0; i<starAmount; i++)
			    		System.out.print("*");
			    	System.out.println();
		    	}
	    	}
		    prevRunTimeId = id;
		    	
	    	prevRunTime = currentTime;
    	}
    }
}