package fi.helsinki.cs.seco.onki.logic.propertyfiddlers;

import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;

import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.RDFList;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.vocabulary.OWL;
import com.hp.hpl.jena.vocabulary.RDF;

import fi.helsinki.cs.seco.onki.logic.IPropertyFiddler;
import fi.helsinki.cs.seco.onki.logic.OnkiResource;
import fi.helsinki.cs.seco.onki.logic.OnkiStatement;

/**
 * Generate owl:equivalentClass relations for objects of owl:equivalentClass relations and
 * YSO aggregate concepts.
 */

public class EquivalentClassPropertyFiddler implements IPropertyFiddler {

    /**
	 * Process objects of owl:equivalentClass relations and
	 * concepts belonging to YSO aggregate concepts.
	 */
	
	public List<OnkiStatement> findBackward(OnkiResource ores) {
		if (!ores.getResource().isResource())
			return null;
		
		List<OnkiStatement> reslist = null;
		
		// case 1) resource is object of a owl:equivalentClass statement
		
		StmtIterator stmti0 = ores.getResource().getModel().listStatements(null, OWL.equivalentClass, ores.getResource());
		if (stmti0.hasNext()) {
			reslist = new LinkedList<OnkiStatement>();
			while (stmti0.hasNext()) {
				Resource res = stmti0.nextStatement().getSubject();
				if (!ores.getResource().getModel().listStatements(ores.getResource(), OWL.equivalentClass, res).hasNext())
					reslist.add(new OnkiStatement(ores, OWL.equivalentClass, res, ores.getLangPreferences(), OnkiStatement.TYPE_INDIRECT | OnkiStatement.TYPE_INVERSE));
			}
		}
		
		// case 2) e.g. YSO aggregateConcept: resource is member of a union list which is object of
		// owl:equivalentClass statement
		
		// TODO: how the union should be modeled?
		for (StmtIterator stmti1 = ores.getResource().getModel().listStatements(null, RDF.first, ores.getResource()); stmti1.hasNext(); ) {
			
			Statement stmt = stmti1.nextStatement();
			
			// find head of the RDF list structure
			Resource head = findListHead((Resource)stmt.getSubject());
	
			// find eq class that is related with a owl:unionOf property to the list
			StmtIterator stmti2 = head.getModel().listStatements(null, OWL.unionOf, head);
			if (!stmti2.hasNext())
				continue;

			Resource eqClass = stmti2.nextStatement().getSubject();
			
			// finally, find the equivalentClass
			StmtIterator stmti3 = eqClass.getModel().listStatements(null, OWL.equivalentClass, eqClass);
			if (!stmti3.hasNext())
				continue;
			
			if (reslist == null)
				reslist = new LinkedList<OnkiStatement>();
			reslist.add(new OnkiStatement(ores, OWL.equivalentClass, stmti3.nextStatement().getSubject(), ores.getLangPreferences(), OnkiStatement.TYPE_INDIRECT | OnkiStatement.TYPE_INVERSE));
		}
		return reslist;
	}
	
	/**
	 * Find head of the list.
	 */
	
	private Resource findListHead(Resource res) {
		StmtIterator stmti = res.getModel().listStatements(null, RDF.rest, res);
		if (!stmti.hasNext())
			return res;
		return findListHead(stmti.nextStatement().getSubject());
	}

	/**
	 * Process YSO aggregate concepts.
	 */
	
	public List<OnkiStatement> findForward(OnkiResource ores, Statement stmt) {
		// case 3) e.g. YSO aggregateConcept: resource is a aggregateConcept having owl:equivalence statement
		// which object is a union list of resources
		if (!stmt.getObject().isResource())
			return null;
		if (!stmt.getPredicate().equals(OWL.equivalentClass))
			return null;
		
		Resource eqObj = (Resource)stmt.getObject();
		Statement unionOfStmt = eqObj.getProperty(OWL.unionOf);
		
		if (unionOfStmt == null)
			return null;

		if (!unionOfStmt.getObject().canAs(RDFList.class))
			return null;
		
		List nodelist = ((RDFList)unionOfStmt.getObject().as(RDFList.class)).asJavaList();
		Iterator nodei = nodelist.iterator();
		List<OnkiStatement> reslist = new LinkedList<OnkiStatement>();
		
		while (nodei.hasNext()) {
			RDFNode node = (RDFNode)nodei.next();
			// TODO: come up with a new property (part of aggregateConcept)
			reslist.add(new OnkiStatement(ores, OWL.equivalentClass, node, ores.getLangPreferences(), OnkiStatement.TYPE_INDIRECT));
		}
		return reslist;
	}
}