package fi.helsinki.cs.seco.onki.servlet;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.ArrayList;

import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.ResourceFactory;

import fi.helsinki.cs.seco.onki.index.IndexResult;
import fi.helsinki.cs.seco.onki.index.IndexResults;
import fi.helsinki.cs.seco.onki.index.IndexResultsMetadata;
import fi.helsinki.cs.seco.onki.service.IOnkiQuery;
import fi.helsinki.cs.seco.onki.service.OnkiHierarchyRelation;
import fi.helsinki.cs.seco.onki.service.OnkiQueryResult;
import fi.helsinki.cs.seco.onki.service.OnkiQueryResults;
import fi.helsinki.cs.seco.onki.service.OnkiQueryResultsMetadata;
import fi.helsinki.cs.seco.onki.service.Statement;
import fi.helsinki.cs.seco.onki.logic.OnkiStatement;
import fi.helsinki.cs.seco.onki.logic.OntologyProcessor;

/**
 * Defines the ONKI API (published as AJAX & SOAP).
 */

public class OnkiQuery implements IOnkiQuery {

	private OntologyProcessor ontologyProcessor = OntologyProcessor.getInstance();

    /**
	 * Search the concept index.
	 */
	
	// maxHits==0 means unlimited number of hits
	public OnkiQueryResults search(String query, String lang, Integer maxHits, Collection<String> types, Collection<String> parents, Collection<String> groups) {
		if (this.isOntologyProcessorSet()) {
			
			if (query == null || query.length() == 0)
				query = "*";
			
			query = query.trim();
			
			if (maxHits == null)
				maxHits = 50;
			
			if ((types == null || types.size() == 0) && ontologyProcessor.getDefaultViewTypes().size() > 0) {
				types = ontologyProcessor.getDefaultViewTypes();
			}
			
			if (types != null && types.size() > 0) {
				HashSet<String> tempTypes = new HashSet<String>();
				for (String type : types)
					if (type != null && type.length()>0)
						tempTypes.add(type);
				types = tempTypes;
			}
			
			// TODO: startFromHit parametrizing?
			//Integer startFromHit = -1;
			Integer startFromHit = 1;
			
			IndexResults results = this.ontologyProcessor.searchConcept(query, lang, types, parents, groups, false, false, maxHits, startFromHit);
			IndexResultsMetadata indexMetadata = results.getMetadata();
			OnkiQueryResultsMetadata metadata = new OnkiQueryResultsMetadata(indexMetadata.getTotalHitsAmount(), indexMetadata.getContainingHitsAmount());
			
			List<OnkiQueryResult> responses = new ArrayList<OnkiQueryResult>();

			for (IndexResult indRes : results.getResults()) {
				OnkiQueryResult qresult = new OnkiQueryResult();
				qresult.setUri(indRes.getUri());
				qresult.setNamespacePrefix(indRes.getUriNamespacePrefix());
				qresult.setTitle(indRes.getPrefLabel());
				qresult.setAltLabel(indRes.getAltLabel());
				qresult.setSerkki(indRes.getPrefLabel());
				responses.add(qresult);
			}
			
			return new OnkiQueryResults(metadata, responses);
			}
		System.out.println("No ontology processor set!");
		return null;
	}
	
    /**
	 * Expand the query.
	 */
	
	public OnkiQueryResults expandQuery(Collection<String> uris, String lang, Integer maxHits, Collection<String> types) {
		if (this.isOntologyProcessorSet()) {
			
			if (maxHits == null)
				maxHits = 50;			
			
			IndexResults results = this.ontologyProcessor.expandQuery(uris, lang, maxHits, types);
			IndexResultsMetadata indexMetadata = results.getMetadata();
			OnkiQueryResultsMetadata metadata = new OnkiQueryResultsMetadata(indexMetadata.getTotalHitsAmount(), indexMetadata.getContainingHitsAmount());
			
			List<OnkiQueryResult> responses = new ArrayList<OnkiQueryResult>();
			
			for (IndexResult indRes : results.getResults()) {
				OnkiQueryResult qresult = new OnkiQueryResult();
				qresult.setUri(indRes.getUri());
				qresult.setNamespacePrefix(indRes.getUriNamespacePrefix());
				qresult.setTitle(indRes.getPrefLabel());
				qresult.setAltLabel(indRes.getAltLabel());
				qresult.setSerkki(indRes.getPrefLabel());
				responses.add(qresult);
			}
			return new OnkiQueryResults(metadata, responses);
			}
		System.out.println("No ontology processor set!");
		return null;
	}
	
    /**
	 * Get the concept hierarchy tree of a concept.
	 */
	
	public Collection<OnkiHierarchyRelation> getConceptTree(String uri, String lang) {
		if (this.isOntologyProcessorSet())
			return this.ontologyProcessor.getConceptTree(uri, this.ontologyProcessor.getLangPreferences(lang));
		System.out.println("No ontology processor set!");
		return null;
	}
	
    /**
	 * Get the properties of a concept.
	 */
	
	public Collection<Statement> getProperties(String uri, Collection<String> propertyUris, String lang) {
		if (this.isOntologyProcessorSet()) {
			if (this.ontologyProcessor.createOnkiResource(uri, null) != null) {
				LinkedList<Statement> stmts = new LinkedList<Statement>();
				HashSet<Property> props = null;
				if (propertyUris != null) {
					props = new HashSet<Property>();
					for (String propertyUri : propertyUris)
						props.add(ResourceFactory.createProperty(propertyUri));
				}
				for (OnkiStatement stmt : this.ontologyProcessor.createOnkiResource(uri, this.ontologyProcessor.getLangPreferences(lang)).getProperties(true, props, true, false)) {
					if (stmt.getObject().isURIResource())
						stmts.add(new Statement(null, null, stmt.getObjectAsURI(), stmt.getObjectAsOnkiResource().getLabel(), stmt.getPredicate().getUri(), stmt.getPredicate().getLabel()));
					else if (stmt.getObject().isLiteral())
						stmts.add(new Statement(stmt.getObjectAsString(), stmt.getLang(), null, null, stmt.getPredicate().getUri(), stmt.getPredicate().getLabel()));
					// TODO: anonymous nodes?
				}
				return stmts;
			}
		}
		System.out.println("No ontology processor set!");
		return null;
	}
	
    /**
	 * Get the languages of labels the concept index contains. 
	 */
	
	public Collection<String> getAvailableLanguages() {
		if (this.isOntologyProcessorSet()) {
			return this.ontologyProcessor.getAvailableLanguages();
		}
		System.out.println("No ontology processor set!");
		return null;
	}
	
    /**
	 * Get the types of concepts the concept index contains.
	 */
	
	public Collection<String> getAvailableTypeUris() {
		if (this.isOntologyProcessorSet()) {
			return this.ontologyProcessor.getAvailableTypeUris();
		}
		System.out.println("No ontology processor set!");
		return null;
	}	
	
    /**
	 * Get the label of a concept.
	 */
	
	public OnkiQueryResult getLabel(String uri, String lang) {
		if (this.isOntologyProcessorSet()) {
			ArrayList<String> langs = new ArrayList<String>();
			langs.add(lang);
			langs.add(""); // if there is no label for the desired language, get the label with no language code 
			if (this.ontologyProcessor.createOnkiResource(uri, langs) != null) {
				String label = this.ontologyProcessor.createOnkiResource(uri, langs).getLabel();
				OnkiQueryResult result = new OnkiQueryResult();
				result.setUri(uri);
				result.setTitle(label);
				return result;
			}
		}
		System.out.println("No ontology processor set!");
		return null;
	}
	
	public void createError(String message) throws Exception {
		throw new Exception(message);
	}
	
    /**
	 * Is the OntologyProcessor set.
	 */
	
	private boolean isOntologyProcessorSet() {
		if (this.ontologyProcessor == null)
			this.ontologyProcessor = OntologyProcessor.getInstance();
		if (this.ontologyProcessor != null) {
			return true;
		}
		else
			return false;
	}
}