package fi.helsinki.cs.seco.onki.servlet;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.Writer;
import java.net.URLDecoder;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;
import java.util.ResourceBundle;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.velocity.Template;
import org.apache.velocity.app.Velocity;
import org.apache.velocity.context.Context;
import org.apache.velocity.servlet.VelocityServlet;

import com.hp.hpl.jena.rdf.model.Resource;

import fi.helsinki.cs.seco.onki.OnkiException;
import fi.helsinki.cs.seco.onki.index.IndexResult;
import fi.helsinki.cs.seco.onki.index.IndexResults;
import fi.helsinki.cs.seco.onki.logic.OnkiConfigurator;
import fi.helsinki.cs.seco.onki.logic.OnkiResource;
import fi.helsinki.cs.seco.onki.logic.OntologyProcessor;
import fi.helsinki.cs.seco.onki.utils.Utf8ResourceBundle;

/**
 * Servlet for handling the HTTP requests of the ONKI SKOS Browser.
 * Handles the URL parameters and forwards queries to OntologyProcessor.
 *
 */

public class OnkiServlet extends VelocityServlet {

    //////////////////////////////////////////////////////////
    //	Constants
    //////////////////////////////////////////////////////////

    public static final String VELOCITY_LOG           = "logs/velocity.log";

    public static final String ONKI_TEMPLATE_DIR      = "WEB-INF/app/templates/";

    public static final String ONKI_ROOT_TEMPLATE     = "onki.vm";

    public static final String ONKI_ERROR_TEMPLATE    = "onki-error.vm";

    public static final String ONKI_EMPTY_TEMPLATE	  = "empty.vm";
    
    public static final String ONKI_MAPPINGS_TEMPLATE = "mappings.vm";

    public static final String ONKI_LOCALE_BUNDLE     = "OnkiLocales";

    private static final String URL_ENCODING          = "UTF-8";

    private static final String DEFAULT_LANG          = "fi";

    private static final String CONTENT_TYPE_RDF	  = "application/rdf+xml";

    private static final String ACCEPT_HEADER		  = "Accept"; 

    /////////////////////////////////////////////////////////
    //	Page identifiers
    /////////////////////////////////////////////////////////
    
    private static final String PAGE_CONCEPT        = "concept";

    private static final String PAGE_SEARCH           = "search";
    
    private static final String PAGE_GROUP_SEARCH           = "group-search";
    
    private static final String PAGE_ONTOLOGY_INFO    = "ontology-info";
    
    private static final String DEFAULT_PAGE          = PAGE_CONCEPT;

    //////////////////////////////////////////////////////////
    //	private variables
    //////////////////////////////////////////////////////////

    private OntologyProcessor processor;
    
    private Log logger = LogFactory.getLog(OnkiServlet.class);

    private Hashtable localeTable = new Hashtable();

    private Hashtable localeDescriptionTable = new Hashtable();

    private Date refreshTime;

    private String defaultLocale;

    private Date startTime;
    private Date cacheExpireTime;
    
    /**
     * Initialization of the servlet.
     */
    
    public void init(ServletConfig config) throws ServletException {
    	super.init(config); // Call velocity init first
    	GregorianCalendar calendar = new GregorianCalendar();
    	this.startTime = calendar.getTime();
    	calendar.add(Calendar.DATE, 1);
    	
    	this.cacheExpireTime = calendar.getTime();
    	try {
    		initProcessor();
    	} catch (Exception e) {
    		e.printStackTrace();
    		throw new ServletException("Error in initializing the processor: "+e);
    	}
    }
    
    //////////////////////////////////////////////////////////
    //	Private helper methods
    //////////////////////////////////////////////////////////

    /**
     * Loads Velocity configuration on init.
     */
    
    @Override
	protected Properties loadConfiguration(ServletConfig config) throws IOException, FileNotFoundException {
        String path = config.getServletContext().getRealPath("/");
        if (path == null) path = "/";
        if (!(path.endsWith("/") || path.endsWith("\\"))) path += "/";

        OntologyProcessor.SERVLET_ROOT_PATH = "" + path;
        refreshTime = new Date(System.currentTimeMillis());

        //	Configure Velocity
        Properties velocityProps = new Properties();
        velocityProps.setProperty(Velocity.FILE_RESOURCE_LOADER_PATH, path + ONKI_TEMPLATE_DIR);
        velocityProps.setProperty(Velocity.COUNTER_INITIAL_VALUE, "0");

        // Force everything to UTF-8. Added by Kim, 5.10.2006

        velocityProps.setProperty(Velocity.INPUT_ENCODING, "UTF-8");
        velocityProps.setProperty(Velocity.OUTPUT_ENCODING, "UTF-8");
        velocityProps.setProperty(Velocity.ENCODING_DEFAULT, "UTF-8");


        // TODO: Before publishing the service, remove these settings
        // During development disable caches!! Kim, 22.2.2007
        velocityProps.setProperty(Velocity.VM_LIBRARY_AUTORELOAD, "true");
        velocityProps.setProperty(Velocity.FILE_RESOURCE_LOADER_CACHE, "false");
        velocityProps.setProperty(Velocity.VM_PERM_ALLOW_INLINE_REPLACE_GLOBAL, "true");
        // ----------------------------------------------------------------------

        return velocityProps;
    }

    /**
     * Errors are thrown through this method.
     */
    
    @Override
	protected void error(HttpServletRequest req, HttpServletResponse res, Exception e)
            throws javax.servlet.ServletException, java.io.IOException {

        Writer result = new StringWriter();
        PrintWriter printWriter = new PrintWriter(result);
        e.printStackTrace(printWriter);
        logger.error("Error occured : " + result.toString());

        if (e instanceof IOException)
			throw (IOException) e;
		else if (e instanceof ServletException) throw (ServletException) e;
		else throw new ServletException("Something very strange happened : " + e);
    }

    /**
	 * The OntologyProcessor handling the ontology data is inited.
	 */

    private void initProcessor() throws Exception {
        // Initialize ontology connection
        if (processor == null) {

            String confFile = OntologyProcessor.SERVLET_ROOT_PATH + OnkiConfigurator.SETTINGS_FILE_APP;
            Configuration conf = null;

            //	Init configuration
            try {
                logger.info("Using root-conf: " + confFile);
                conf = OnkiConfigurator.getConfiguration(confFile);
                String values = "";
                for (Iterator i = conf.getKeys(); i.hasNext(); ) {
                	String key = (String)i.next();
                	values += key + "=";
                	for (Iterator iter = conf.getList(key).iterator(); iter.hasNext(); ) {
                		values += iter.next();
                		if (iter.hasNext())
                			values += ",";
                	}        	
                	values += ";\n";
                }
                logger.info("configuration: " + values);
            } catch (IOException ioe) {
                logger.error("Could not init Configuration (IOException): " + ioe.getStackTrace());
                throw ioe;
            } catch (ConfigurationException cfe) {
                logger.error("Could not init Configuration (ConfigurationException): " + cfe.getStackTrace());
                throw cfe;
            }
            
            //	Init processor
            try {
                processor = OntologyProcessor.getInstance(conf);
            } catch (Throwable t) {
                logger.error("Could not init processor : " + t);
                throw new Exception("Could not init processor : " + t);
            }

            //	Initialize Locales

            try {
                initLocales(processor.getConfig());
            } catch (Exception e) {
                logger.error("Could not init localization : " + e.getStackTrace());
            }

            processor.setStartTime(this.startTime);
            processor.setCacheExpireTime(this.cacheExpireTime);
            
            //	Load ontologies
            try {
                processor.loadOntologies();
                logger.info("All files attached.");
                processor.indexOntologies();
                logger.info("Ontologies were indexed.");
                processor.generateDefaultDirectoryViews();
                logger.info("Default ontology directory views were generated.");
            } catch (IOException ioe) {
                logger.error(getClass() + " : Error initializing ontologys (IOException)" + ioe.getMessage());
                throw ioe;
            } catch (OnkiException oe) {
                logger.error(getClass() + " : Error initializing ontologys (OnkiException) " + oe.getMessage());
                throw oe;
            } catch (Exception e) {
                logger.error(getClass() + " : Error initializing ontologys (Exception) " + e.getStackTrace());
            	e.printStackTrace();
                throw e;
            }

            logger.info("ONKI is up and running.");
        }
    }

    /**
	 * The locales are inited.
	 */

    private void initLocales(Configuration conf) throws Exception {
    	Collection localeCol = conf.getList(OnkiConfigurator.ELEM_VISUALIZATION_LOCALE);
        defaultLocale = conf.getString(OnkiConfigurator.ELEM_VISUALIZATION_LOCALES_DEFAULT);
        if (!localeCol.isEmpty()) {
            for (int i = 0; i < localeCol.size(); i++) {
                String localeDescription = conf.getString(OnkiConfigurator.ELEM_VISUALIZATION_LOCALE + "(" + i + ")");
                String localeId = conf.getString(OnkiConfigurator.ELEM_VISUALIZATION_LOCALE + "(" + i + ")."
                        + OnkiConfigurator.ELEM_VISUALIZATION_LOCALE_ID);
                ResourceBundle bundle = Utf8ResourceBundle.getBundle(ONKI_LOCALE_BUNDLE, new Locale(localeId));
                localeTable.put(localeId, bundle);
                localeDescriptionTable.put(localeId, localeDescription);
                logger.info("Found :" + localeId + ": " + localeDescription);
            }

            //	Return default language if not found
            if (defaultLocale == null || !localeTable.containsKey(defaultLocale)) defaultLocale = DEFAULT_LANG;
            logger.info("Locales inited");
        } else logger.error("Locales NOT inited");

    }

    /**
	 * Get a ResourceBundle for a locale. 
	 */
    
    private ResourceBundle getLocaleBundle(String lang) {
        if (lang == null)
        	lang = this.defaultLocale;
        else if (!localeTable.containsKey(lang)) lang = this.defaultLocale;
        return (ResourceBundle) localeTable.get(lang);
    }
    
    //////////////////////////////////////////////////////////
    //	Public methods for handling requests
    //////////////////////////////////////////////////////////

    /**
	 * Get the start time of OnkiServlet as the last modified timestamp.
	 * Proxy/browser cache implementations can use this.
	 */
    
    public long getLastModified(HttpServletRequest request) {
    	return this.startTime.getTime();
    }
    
    /**
     * Main routine to handle a request. Called by VelocityServlet.
     *
     * @param ctx Velocity Context object to be filled with data. Will be used
     *            for rendering this template.
     * @return Template to be used for request
     */
    @Override
    
	public Template handleRequest(HttpServletRequest request, HttpServletResponse response, Context ctx)
            throws Exception {

        response.setCharacterEncoding("UTF-8");
        request.setCharacterEncoding("UTF-8");
        
        //response.setHeader("cache-control", "x");
        
        // comment out for dev
        /*SimpleDateFormat df = new SimpleDateFormat("EEE, d MMM yyyy HH:mm:ss Z", Locale.US);
        response.setHeader("Last-Modified", df.format(this.startTime));
        response.setHeader("Expires", df.format(this.cacheExpireTime));*/
        
        //response.setContentType("application/xhtml+xml");        

    	Map<String,String[]> requestParams = new HashMap<String,String[]>(request.getParameterMap());
    	String baseurl = request.getRequestURL().toString();
        OnkiURLGenerator onkiUrl = new OnkiURLGenerator(baseurl, requestParams, URL_ENCODING);
        
        // Handle default view
        {
        	String[] queryType = onkiUrl.getCurrentUrl().getQueryType();
	        String activeConcept = onkiUrl.getCurrentUrl().getConcept();
	        
	        // return concept as rdf
	        String acceptHeader = request.getHeader(ACCEPT_HEADER);
	        
	        if ((requestParams.containsKey(OnkiURLGenerator.PARAM_RDF) || (acceptHeader != null && acceptHeader.equals(CONTENT_TYPE_RDF))) && activeConcept != null) {
	        	response.setContentType(CONTENT_TYPE_RDF);
	        	this.processor.getResourceAsRDF(activeConcept, response.getOutputStream());
	        	response.getOutputStream().flush();
	        	return getTemplate(ONKI_EMPTY_TEMPLATE);
	        }
	        
	        // TEMPORARY PLACE FOR DEMO FOR CO-OCCURENCE MAPPING GENERATION UI, Jouni 15.9.2008
	        if (requestParams.containsKey(OnkiURLGenerator.PARAM_MAPPINGS)) {
	        	ctx.put("mappings", this.processor.getYsoCoOccurences(this.getLangPreferences(onkiUrl.getCurrentUrl().getLang())));
	        	return getTemplate(ONKI_MAPPINGS_TEMPLATE);
	        }
	        
	        if (activeConcept == null || activeConcept.trim().length() == 0) {
	        	activeConcept = processor.getDefaultViewConcept();
	        	if (activeConcept != null) {
	    			requestParams.put(OnkiURLGenerator.PARAM_CONCEPT, new String[]{activeConcept});
	        	}
	        }
	        
	        if (queryType == null) {
	        	if (processor.getDefaultViewTypes().size() > 0)
	        		requestParams.put(OnkiURLGenerator.PARAM_QUERY_TYPE, processor.getDefaultViewTypes().toArray(new String[]{}));
	        }
	        else if (queryType.length == 1 && queryType[0].trim().length() == 0) {
	        	queryType[0] = "";
	        	requestParams.put(OnkiURLGenerator.PARAM_QUERY_TYPE, queryType);
	        }

	        onkiUrl = new OnkiURLGenerator(baseurl, requestParams, URL_ENCODING);
        }

        // for emphasizing the selected search alphabet
        List alphabets = processor.getAlphabets();
        
        String queryStr = onkiUrl.getCurrentUrl().getQuery();
        if (queryStr == null)
        	queryStr = "";
        else {
            try {
                queryStr = URLDecoder.decode(queryStr, "UTF-8");
            } catch (Exception e) {
                logger.error("<handleRequest>" + e);
            }
         	queryStr = queryStr.trim();
        }

        if (queryStr.length()==0)
        	if (processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_ALPHA_INDEX_DEFAULT))
        		queryStr = processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_ALPHA_INDEX_DEFAULT);
        	else
        		queryStr = (String)alphabets.get(1);
        queryStr = queryStr.toLowerCase();
        
        if (queryStr.length()==1)
        	ctx.put("activeAlphabet", queryStr);

        String groupQueryStr = onkiUrl.getCurrentUrl().getGroupQuery();
        if (groupQueryStr == null)
        	groupQueryStr = "";
        else {
            try {
            	groupQueryStr = URLDecoder.decode(groupQueryStr, "UTF-8");
            } catch (Exception e) {
                logger.error("<handleRequest>" + e);
            }
            groupQueryStr = groupQueryStr.trim();
        }

        if (groupQueryStr.length()==0)
        	if (processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_GROUP_DEFAULT_QUERY))
        		groupQueryStr = processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_GROUP_DEFAULT_QUERY);
        	else
        		groupQueryStr = (String)alphabets.get(1);
        groupQueryStr = groupQueryStr.toLowerCase();
        
        if (groupQueryStr.length()==1)
        	ctx.put("activeGroupAlphabet", groupQueryStr);
        
        onkiUrl.setDefault(onkiUrl.PARAM_LANG, this.defaultLocale);
        onkiUrl.setDefault(onkiUrl.PARAM_PAGE, DEFAULT_PAGE);
        ctx.put("onkiurl", onkiUrl);
        
       // in concept fetch mode the "host" website's url needs to be known
       ctx.put("referer", onkiUrl.getCurrentUrl().getReferer());

        List langPreferences = this.getLangPreferences(onkiUrl.getCurrentUrl().getLang());
        ctx.put("langPreferences", langPreferences);
        
        if (processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_TYPE_RESTRICTION_ENABLED) && !processor.getConfig().getBoolean(OnkiConfigurator.ELEM_CLASSES_TYPE_RESTRICTION_ENABLED)) {
        	ctx.put("type-restriction-menu", Boolean.FALSE);
        }        
        else {
        	ctx.put("type-restriction-menu", Boolean.TRUE);
        	Collection<OnkiResource> resourceTypes = this.processor.getAvailableTypes(langPreferences);
        	ctx.put("availableResourceTypes", resourceTypes);
        }
        	
        //////////////////////////////////////////////////////
        //	Resolve language

        ctx.put("locale", getLocaleBundle(onkiUrl.getCurrentUrl().getLang()));

        ctx.put("defaultLang", this.defaultLocale);

        //////////////////////////////////////////////////////
        //	Set localization
        
        ctx.put("localesTable", localeDescriptionTable);

        //////////////////////////////////////////////////////
        //	Resolve the request page

        String page = onkiUrl.getCurrentUrl().getPage();
        
        if (processor.getConfig().getBoolean(OnkiConfigurator.ELEM_MULTIONKI_ENABLED)) {
        	ctx.put("multionki", Boolean.TRUE);
        	if (processor.getConfig().containsKey(OnkiConfigurator.ELEM_MULTIONKI_EMPHASIZING) && processor.getConfig().getBoolean(OnkiConfigurator.ELEM_MULTIONKI_EMPHASIZING))
        		ctx.put("emphasizing", Boolean.TRUE );
        }
		else ctx.put("multionki", Boolean.FALSE);
        
        if (processor.getConfig().getBoolean(OnkiConfigurator.ELEM_ONTOLOGY_INFO_PAGE_ENABLED)) {
        	ctx.put("ontology-info", Boolean.TRUE);
        	ctx.put("ontology-info-url", processor.getConfig().getString(OnkiConfigurator.ELEM_ONTOLOGY_INFO_PAGE));
        }
		else ctx.put("ontology-info", Boolean.FALSE);     
        
        if (processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_GROUP_ENABLED) && processor.getConfig().getBoolean(OnkiConfigurator.ELEM_CLASSES_GROUP_ENABLED)) {
        	ctx.put("group-search", Boolean.TRUE);
        	if (processor.getConfig().getBoolean(OnkiConfigurator.ELEM_CLASSES_GROUP_SEARCHABLE))
        		ctx.put("group-searchable", Boolean.TRUE);;
        }
        
       	ctx.put("google-maps-key", processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_MAP_PROPERTIES_GOOGLE_MAPS_KEY));
        
       	ctx.put("pointCircleRadius", processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_MAP_PROPERTIES_POINT_PROPERTY_RADIUS));

        String template = ONKI_ROOT_TEMPLATE;
        String handler = page;

        boolean showHeader = true;
        boolean showNavigation = true;
        int maxhits = -1;

        if (handler == null)
        	handler = DEFAULT_PAGE;

        if (page != null && (page.equals(PAGE_SEARCH) || page.equals(PAGE_GROUP_SEARCH))) showHeader = false;

        ////////////////////////////////////////////////////
        //    	Go to page-spesific handler

        ctx.put("ontology-label", this.generateOnkiLabel(langPreferences, true));
        ctx.put("ontology-sublabel", this.generateOnkiLabel(langPreferences, false));
        ctx.put("alphabetList", processor.getAlphabets());
        ctx.put("page", page);
        ctx.put("showHeader", showHeader);
        ctx.put("showNavigation", showNavigation);
        ctx.put("mainview", page+"-view.vm");

        if (handler.equals(PAGE_CONCEPT))
			this.handleConceptHierarchyView(request, response, ctx);
		else if (handler.equals(PAGE_SEARCH))
			this.handleConceptSearchView(request, response, ctx, maxhits, true);
		else if (handler.equals(PAGE_GROUP_SEARCH))
			this.handleConceptSearchView(request, response, ctx, maxhits, false);
		else if (handler.equals(PAGE_ONTOLOGY_INFO)) ctx.put("stats", processor.getOntologyStatistics());

        return getTemplate(template);
    }

    /**
     * Handle autocompletion search.
     * 
     * @param request
     * @param response
     * @param ctx
     * @throws Exception
     */
    public void handleConceptSearchView(HttpServletRequest request, HttpServletResponse response, Context ctx, int maxhits, boolean conceptSearch)
    throws Exception {

		List langPreferences = (List)ctx.get("langPreferences");
		OnkiURLGenerator onkiUrl   =  (OnkiURLGenerator)ctx.get("onkiurl");
		
		List alphabets = processor.getAlphabets();
		
		OnkiURL url = onkiUrl.getCurrentUrl();
		String queryStr;
		
		if (conceptSearch)
			queryStr = url.getQuery();
		else
			queryStr = url.getGroupQuery();
		 
		if (queryStr == null)
			queryStr = "";
		else {
		    try {
		        queryStr = URLDecoder.decode(queryStr, "UTF-8");
		    } catch (Exception e) {
		        logger.error("<handleConceptSearchView>" + e);
		    }
		 	queryStr = queryStr.trim();
		}
		
		if (queryStr.length()==0)
			if (!conceptSearch && processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_GROUP_DEFAULT_QUERY))
				queryStr = processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_GROUP_DEFAULT_QUERY);
			else {
				if (processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_ALPHA_INDEX_DEFAULT))
	        		queryStr = processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_ALPHA_INDEX_DEFAULT);
	        	else
	        		queryStr = (String)alphabets.get(1);
			}
		queryStr = queryStr.toLowerCase();
		
		LinkedList<String> types = null;
		LinkedList<String> parents = null;
		LinkedList<String> groups = null;
		
		boolean isGroup = false;
		
		if (conceptSearch) {
			if (url.getQueryParent() != null) {
				parents = new LinkedList<String>();
				for (String parent : url.getQueryParent())
					if (parent != null && parent.length()>0)
						parents.add(parent);
				if (parents.size() == 0)
					parents = null;
			}
			if (url.getQueryType() != null) {
				types = new LinkedList<String>();
				for (String type : url.getQueryType())
					if (type != null && type.length()>0)
						types.add(type);
				if (types.size() == 0)
					types = null;
			}
			if (url.getQueryGroup() != null) {
				groups = new LinkedList<String>();
				for (String group : url.getQueryGroup())
					if (group != null && group.length()>0)
						groups.add(group);
				if (groups.size() == 0)
					groups = null;
			}
		}
		else {
			isGroup = true;
			ctx.put("groupquery", true);
			if (!processor.getConfig().getBoolean(OnkiConfigurator.ELEM_CLASSES_GROUP_SEARCHABLE))
				queryStr = "*";
		}
			
		IndexResults results = processor.searchConcept(queryStr, onkiUrl.getCurrentUrl().getLang(), types, parents, groups, isGroup, false, -1, -1);
		Collection<IndexResult> result = results.getResults();
		
		ctx.put("queryresult", result);
	}
    
    /**
     * Handle the concept hierarchy view.
     * 
     * @param request
     * @param response
     * @param ctx
     */
    
    public void handleConceptHierarchyView(HttpServletRequest request, HttpServletResponse response, Context ctx) {
        OnkiURLGenerator onkiUrl   =  (OnkiURLGenerator)ctx.get("onkiurl");

        String activeConcept       = onkiUrl.getCurrentUrl().getConcept();
        List langPreferences       = (List)ctx.get("langPreferences");

        if (activeConcept != null && activeConcept.trim().equals("")) activeConcept = null;

        Resource activeResource = processor.getConcept(activeConcept);
        OnkiResource ores = null;

        if (activeResource != null && processor.containsResource(activeResource)) {
        	ores = new OnkiResource(activeResource, langPreferences);
        	processor.addConceptHierarchy(ores, langPreferences);
        	
        	// for the UI to be able to show if a child/sibling of the current resource has children, might be slooow
        	// comment the following 4 lines if too slow
        	/*for (OnkiResource child : ores.getChilds())
        		processor.addChildConcepts(child, langPreferences);
        	for (OnkiResource sibling : ores.getSiblings())
        		processor.addChildConcepts(sibling, langPreferences);*/
        }

        ctx.put("activeConcept", ores);
        
        String[] queryGroups = onkiUrl.getCurrentUrl().getQueryGroup();
        if (queryGroups != null && queryGroups.length > 0) {
        	Resource groupResource = processor.getConcept(queryGroups[0]);
	        OnkiResource groupOres = null;
	        
	        // TODO: multiple query groups
	        if (groupResource != null && processor.containsResource(groupResource)) {
	        	groupOres = new OnkiResource(groupResource, langPreferences);
	        	ctx.put("queryGroup", groupOres);
	        }
        }
    }

    /**
	 * Get the language preferences for user interface.
	 */
    
    private List getLangPreferences(String lang) {
    	return this.processor.getLangPreferences(lang);
    }

    /**
	 * Generate the top heading to the user interface.
	 */
    
    private String generateOnkiLabel(List<String> langPreferences, boolean mainLabel) {
    	String label = null;
    	
    	String labelElement;
    	String labelLangElement;
    	
    	if (mainLabel) {
    		labelElement = OnkiConfigurator.ELEM_CLASSES_ONTOLOGY_LABEL;
    		labelLangElement = OnkiConfigurator.ELEM_CLASSES_ONTOLOGY_LABEL_LANG;
    	}
    	else {
    		labelElement = OnkiConfigurator.ELEM_CLASSES_ONTOLOGY_SUBLABEL;
    		labelLangElement = OnkiConfigurator.ELEM_CLASSES_ONTOLOGY_SUBLABEL_LANG;
    	}
    	
	    // First check if the label is defined in conf file
    	
    	List ontologyLabelLanguages = processor.getConfig().getList(labelLangElement);
    	List ontologyLabels = processor.getConfig().getList(labelElement);
    	
	    if (ontologyLabels.size() > 0) if (ontologyLabels.size() == ontologyLabelLanguages.size()) {
			Iterator labelIte = ontologyLabels.iterator();
			for (Iterator labelLanguagesIte = ontologyLabelLanguages.iterator(); labelLanguagesIte.hasNext(); )
				if (labelLanguagesIte.next().equals(langPreferences.get(0))) {
					label = (String)labelIte.next();
					break;
				}
				else
					labelIte.next();
			if (label == null)
				label = processor.getConfig().getString(labelElement); // There isn't a label for current language, so we'll use the first label
		}
		else  // There isn't a language attribute for every label in conf file, so we'll use the first label
			label = processor.getConfig().getString(labelElement); // There isn't a label for current language, so we'll use the first label

	    // If the label isn't defined in conf file, get it from ontology resource
	    if (mainLabel && label == null) {
	    Resource ontologyResource = processor.getOntologyResource();
	    OnkiResource ores = null;
	        if (ontologyResource != null && processor.containsResource(ontologyResource)) {
	        	ores = new OnkiResource(ontologyResource, langPreferences);
	    		label = ores.getLabel(langPreferences.get(0));
	        }
	    }
	    // If the label isn't defined in the ontology resource, get it from ConceptScheme(s)
	    if (mainLabel && label == null) {
	    	Collection<Resource> conceptSchemes = processor.getConceptSchemes();
	    	if (!conceptSchemes.isEmpty()) {
	    		label = "";
	    		for (Iterator<Resource> iter = conceptSchemes.iterator(); iter.hasNext(); ) {
	    			Resource res = iter.next();
	    			OnkiResource ores = new OnkiResource(res, langPreferences);
	    			label += ores.getLabel(langPreferences.get(0));
	    			if (iter.hasNext())
	    				label += ", ";
	    		}
	    	}
	    }
	    return label;
    }
}