package fi.helsinki.cs.seco.onki.servlet;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.restlet.Context;
import org.restlet.data.Form;
import org.restlet.data.MediaType;
import org.restlet.data.Parameter;
import org.restlet.data.Request;
import org.restlet.data.Response;
import org.restlet.ext.json.JsonRepresentation;
import org.restlet.resource.Representation;
import org.restlet.resource.ResourceException;
import org.restlet.resource.Variant;

import fi.helsinki.cs.seco.onki.service.OnkiHierarchyRelation;
import fi.helsinki.cs.seco.onki.service.OnkiQueryResult;
import fi.helsinki.cs.seco.onki.service.OnkiQueryResults;

/**
 * Represents a result set of concept search.
 */

public class SearchResource extends BaseResource {
	String queryString;
	String lang;
	Integer maxHits;
	Integer startFromHit;
	Collection<String> types;
	Collection<String> excludedTypes;
	Collection<String> parents;
	Collection<String> excludedParents;
	Collection<String> groups;
	Collection<String> excludedGroups;
	
	int parentAmount = 0;
	int childAmount = 0;
	int siblingAmount = 0;
	
	OnkiQueryResults results;

	Map<String, Collection<OnkiHierarchyRelation>> conceptHierarchies;
	
	boolean autocompletion = true;
	boolean multipleHitsForConcept = false;
	
	public SearchResource(Context context, Request request, Response response) {
		super(context, request, response);
			   
		Form query = getRequest().getResourceRef().getQueryAsForm();
		 
		for (Parameter param : query) {
			if (param.getName().equals("q"))
				this.queryString = param.getValue();
			else if (param.getName().equals("l")) {
				if (param.getValue() != null)
					this.lang = param.getValue();
				else
					this.lang = "";
			}
			else if (param.getName().equals("h"))
				// TODO: handle errors
				this.maxHits = new Integer(param.getValue());
			else if (param.getName().equals("s"))
				// TODO: handle errors
				this.startFromHit = new Integer(param.getValue());
			else if (param.getName().equals("qt")) {
				if (this.types == null)
					this.types = new LinkedList<String>();
				this.types.add(param.getValue());
			}
			else if (param.getName().equals("eqt")) {
				if (this.excludedTypes == null)
					this.excludedTypes = new LinkedList<String>();
				this.excludedTypes.add(param.getValue());
			}
			else if (param.getName().equals("qp")) {
				if (this.parents == null)
					this.parents = new LinkedList<String>();
				this.parents.add(param.getValue());
			}
			else if (param.getName().equals("eqp")) {
				if (this.excludedParents == null)
					this.excludedParents = new LinkedList<String>();
				this.excludedParents.add(param.getValue());
			}
			else if (param.getName().equals("qg")) {
				if (this.groups == null)
					this.groups = new LinkedList<String>();
				this.groups.add(param.getValue());
			}
			else if (param.getName().equals("eqg")) {
				if (this.excludedGroups == null)
					this.excludedGroups = new LinkedList<String>();
				this.excludedGroups.add(param.getValue());
			}
			else if (param.getName().equals("ac"))
				// TODO: handle errors
				this.autocompletion = new Boolean(param.getValue());
			else if (param.getName().equals("mh"))
				// TODO: handle errors
				this.multipleHitsForConcept = new Boolean(param.getValue());
			else if (param.getName().equals("p")) {
				if (param.getValue().equals("*"))
					this.parentAmount = -1;
				else
					this.parentAmount = new Integer(param.getValue());
			}
			else if (param.getName().equals("c")) {
				if (param.getValue().equals("*"))
					this.childAmount = -1;
				else
					this.childAmount = new Integer(param.getValue());
			}
			else if (param.getName().equals("si")) {
				if (param.getValue().equals("*"))
					this.siblingAmount = -1;
				else
					this.siblingAmount = new Integer(param.getValue());
			}
		}
		if (queryString == null)
			queryString = "*";

		this.results = search(this.queryString, this.lang, this.maxHits, this.startFromHit, this.types, this.parents, this.groups, this.excludedTypes, this.excludedParents, this.excludedGroups, this.autocompletion, this.multipleHitsForConcept);

		if (parentAmount != 0 || childAmount != 0 || siblingAmount != 0) {
			this.conceptHierarchies = new HashMap<String, Collection<OnkiHierarchyRelation>>();
			for (OnkiQueryResult result : this.results.getResults()) {
				Collection<OnkiHierarchyRelation> hierarchy = this.getConceptHierarchy(result.getUri(), lang, parentAmount, childAmount, siblingAmount);
				if (hierarchy != null && hierarchy.size() > 0)
					this.conceptHierarchies.put(result.getUri(), hierarchy);
			}
		}
		
		getVariants().add(new Variant(MediaType.APPLICATION_JSON));
	}
	   
	/** 
	 * Returns a full representation for a given variant. 
	 */  
	@Override  
	public Representation represent(Variant variant) throws ResourceException {
		JSONObject object = new JSONObject();
		JSONObject metadata = new JSONObject();
		JSONArray results = new JSONArray();
	 
		try {
			metadata.put("containingHitsAmount", this.results.getMetadata().getContainingHitsAmount());
			metadata.put("moreHitsAmount", this.results.getMetadata().getMoreHitsAmount());
			for (OnkiQueryResult result : this.results.getResults()) {
				JSONObject resultObj = new JSONObject();
				resultObj.put("uri", result.getUri());
				resultObj.put("label", result.getTitle());
				resultObj.put("altLabel", result.getAltLabel());
				resultObj.put("serkki", result.getSerkki());
				resultObj.put("namespacePrefix", result.getNamespacePrefix());
				
				if (parentAmount != 0 || childAmount != 0 || siblingAmount != 0) {
					JSONArray hierarchy = new JSONArray();
					if (this.conceptHierarchies != null && this.conceptHierarchies.containsKey(result.getUri()))
						for (OnkiHierarchyRelation relation : this.conceptHierarchies.get(result.getUri())) {
							JSONObject hierachyRelation = new JSONObject();
							try {
								hierachyRelation.put("conceptUri", relation.getConceptUri());
								hierachyRelation.put("label", relation.getLabel());
								hierachyRelation.put("parentUri", relation.getParentUri());
								hierachyRelation.put("indent", relation.getIndent());
							}
							catch (JSONException e) {
								e.printStackTrace();
							}
							hierarchy.put(hierachyRelation);
						}
					resultObj.put("hierarchy", hierarchy);
				}

				results.put(resultObj);
			}
			object.put("metadata", metadata);
			object.put("results", results);
		}
		catch (JSONException e) {
			;
		}

		JsonRepresentation representation = new JsonRepresentation(object);
		representation.setModificationDate(getLastModified());
		representation.setExpirationDate(getExpires());
		this.finalize();
	    return representation;
	 }  
}