package fi.seco.semweb.infrastructure.jena;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.apache.commons.collections15.BidiMap;
import org.apache.commons.collections15.bidimap.DualHashBidiMap;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.hp.hpl.jena.db.DBConnection;
import com.hp.hpl.jena.db.IDBConnection;
import com.hp.hpl.jena.db.ModelRDB;
import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.ontology.OntModelSpec;
import com.hp.hpl.jena.rdf.listeners.ObjectListener;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.ResourceFactory;
import com.hp.hpl.jena.shared.PrefixMapping;
import com.hp.hpl.jena.tdb.TDBFactory;

import fi.seco.semweb.util.map.primitive.ObjectLongHashMap;
import fi.seco.semweb.util.sql.KeepAliveConnection;
import fi.seco.semweb.util.sql.NonExceptingConnection;
import fi.seco.semweb.util.sql.NonExceptingResultSet;
import fi.seco.semweb.util.sql.NonExceptingStatement;

public class JenaService {

	private static Log log = LogFactory.getLog(JenaService.class);

	private PrefixMapping globalPrefixMap = PrefixMapping.Factory.create();

	private ObjectLongHashMap<Object> lastModified = new ObjectLongHashMap<Object>();
	private BidiMap<String,Object> models = new DualHashBidiMap<String,Object>();
	private Set<IJenaUrlModelSource> updateWatch = new HashSet<IJenaUrlModelSource>();

	private boolean eagerLoad=false;

	private List<JenaModelSource> sources;

	private Map<Model,Collection<IModelReloadEventListener>> eventListeners = new HashMap<Model,Collection<IModelReloadEventListener>>();

	public void addGlobalPrefixesToModel(Model m) {
		m.withDefaultMappings(globalPrefixMap);
	}
	
	private Model addModelFromComboTDBSource(JenaComboTDBModelSource js) {
		long dirLastModified = js.getLastModified();
		Model m = TDBFactory.createModel(js.getLocation());
		putModel(js.getKey(), m);
		lastModified.put(js.getKey(), dirLastModified);
		boolean reload = false;
		for (String url : js.getUrls())
		try {
			long lm = new URL(url).openConnection().getLastModified();
			log.info(url+" last modified "+new Timestamp(lm));
			if (lm>getLastModified(js.getKey())) reload=true;
		} catch (MalformedURLException e) {
			log.error("Malformed URL " +url+" when reading Jena model "+js.getKey(),e);
		} catch (IOException e) {
			log.error("Read from URL "+url+" failed when reading Jena model "+js.getKey(),e);
		}
		
		if (reload) {
			log.info("At least one URL is fresher than TDB, reloading");
			m.removeAll();
			readModelFromUrlSource(new JenaUrlModelSource(js.getKey(),js.getUrls()));
		}
		updateWatch.add(js);
		return m;
	}
	
	private Model addModelFromComboSource(JenaComboModelSource js) {
		IDBConnection conn = getConnection(js);
		ModelRDB m;
		if (conn.containsModel(js.getModel())) m = addModelFromDBSource(js);
		else {
			log.info("model "+js.getModel()+" doesn't exist in the database, creating it.");
			m = fixDBModel(js);
			putModel(js.getKey(),m);
			lastModified.put(js.getKey(),0);
		}
		boolean reload=false;
		for (String url : js.getUrls())
			try {
				long lm = new URL(url).openConnection().getLastModified();
				log.info(url+" last modified "+new Timestamp(lm));
				if (lm>getLastModified(js.getKey())) reload=true;
			} catch (MalformedURLException e) {
				log.error("Malformed URL " +url+" when reading Jena model "+js.getKey(),e);
			} catch (IOException e) {
				log.error("Read from URL "+url+" failed when reading Jena model "+js.getKey(),e);
			}
		if (reload) {
			log.info("At least one URL is fresher than database, reloading");
			m.removeAll();
			m.setDoDuplicateCheck(false); // Jouni, 28.10.2008
			//readModelFromUrlSource(new JenaUrlModelSource(js.getKey(),js.getUrls(),js.getFormat()));
			readModelFromUrlSource(new JenaUrlModelSource(js.getKey(),js.getUrls()));
			try {
				if (!js.getType().equals("Derby")) // Jouni, 28.10.2008
					conn.getConnection().createStatement().executeUpdate("UPDATE jena_graph SET LastModified=CURRENT_TIMESTAMP WHERE name='"+js.getModel()+"'");
			} catch (SQLException e) {
				log.error("Couldn't update last modified field in database",e);
			}
		}
		updateWatch.add(js);
		return m;
	}

	private Model addModelFromSource(IJenaModelSource js) {
		if (js instanceof JenaComboModelSource) return addModelFromComboSource((JenaComboModelSource) js);
		else if (js instanceof JenaUrlModelSource) return addModelFromUrlSource((JenaUrlModelSource) js);
		else if (js instanceof JenaModelCombinationSource) return addModelFromCombinationSource((JenaModelCombinationSource) js);
		else if (js instanceof JenaComboTDBModelSource) return addModelFromComboTDBSource((JenaComboTDBModelSource) js);
		else return addModelFromDBSource((JenaDBModelSource) js);
	}
	
	private Model addModelFromCombinationSource(JenaModelCombinationSource source) {
		OntModel m = ModelFactory.createOntologyModel(OntModelSpec.OWL_MEM);
		log.info("Creating model "+source.getKey()+" from models:");
		for (String key : source.getKeys()) {
			log.info(" * "+key);
			if (getModel(key)!=null)
				m.addSubModel(getModel(key));
			else log.error("Model "+key+" not defined, not adding to combination model!");
		}
		putModel(source.getKey(),m);
		log.info("done.");
		return m;
	}

	private ModelRDB addModelFromDBSource(JenaDBModelSource js) {
		log.info("Loading model "+js.getKey()+" from DB source: "+js);
		//IDBConnection conn = getConnection(js);
		ModelRDB m = fixDBModel(js);
		m.setQueryOnlyAsserted(true);
		putModel(js.getKey(),m);
		long lm = js.getLastModified();
		log.info("Database last modified: "+new Timestamp(lm));
		lastModified.put(js.getKey(),lm);
		/*try {
			Connection sqconn = conn.getConnection();
			ResultSet rs = sqconn.createStatement().executeQuery("select LastModified from jena_graph where Name='"+js.getModel()+"'");
			rs.next();
			log.info("Database last modified: "+rs.getTimestamp(1));
			lastModified.put(js.getKey(),rs.getTimestamp(1).getTime());
		} catch (SQLException e) {
			log.error("Couldn't get last modified timestamp from SQL",e);
		}*/
		return m;
	}

	private Model addModelFromUrlSource(JenaUrlModelSource js) {
		Model m = ModelFactory.createDefaultModel();
		putModel(js.getKey(),m);
		lastModified.put(js.getKey(),0);
		return readModelFromUrlSource(js);
	}

	private Model readModelFromUrlSource(IJenaUrlModelSource js) {
		Model m = (Model) models.get(js.getKey());
		for (String url : js.getUrls()) {
			log.info("Loading url source "+url+" into model "+js.getKey());
			try {
				URLConnection uc = new URL(url).openConnection();
				InputStream ic = uc.getInputStream();
				//m.read(ic,null);
				//m.read(ic,null,"TURTLE");
				//m.read(ic,null,js.getFormat());
				m.read(ic,null,parseType(url));
				//m.read(ic,null,"N3");
				ic.close();
				log.info("done.");
				//lastModified.put(js.getKey(),uc.getLastModified()); // removed, Jouni, 13.3.2009
			} catch (MalformedURLException e) {
				log.error("Malformed URL " +url+" when reading Jena model "+js.getKey(),e);
			} catch (IOException e) {
				log.error("Read from URL "+url+" failed when reading Jena model "+js.getKey(),e);
			}
		}
		long lm = js.getLastModified();
		lastModified.put(js.getKey(),lm);
		updateWatch.add(js);
		return m;
	}

	@SuppressWarnings("unchecked")
	public void copyMappings(PrefixMapping from, PrefixMapping to) {
		for (Object mpmo : from.getNsPrefixMap().entrySet()) {
			Map.Entry<String,String> mpm = (Entry<String, String>) mpmo;
			if (to.getNsPrefixURI(mpm.getKey())==null && to.getNsURIPrefix(mpm.getValue())==null) to.setNsPrefix(mpm.getKey(),mpm.getValue());
			else if (!mpm.getValue().equals(to.getNsPrefixURI(mpm.getKey())) || !mpm.getKey().equals(to.getNsURIPrefix(mpm.getValue()))) log.warn("Prefix/namespace mapping "+mpm.getKey()+"="+mpm.getValue()+" differs from previous definition. Previously "+mpm.getValue()+"="+to.getNsURIPrefix(mpm.getValue())+", "+mpm.getKey()+"="+to.getNsPrefixURI(mpm.getKey())+". PrefixMapping not modified.");
		}
	}

	public Model createDefaultModel() {
		Model m = ModelFactory.createDefaultModel();
		m.setNsPrefixes(globalPrefixMap);
		return m;
	}

	public IDBConnection getConnection(String modelKey) {
		for (IJenaModelSource jSource : sources)
			if(jSource.getKey().equalsIgnoreCase(modelKey) && jSource instanceof JenaDBModelSource) return getConnection((JenaDBModelSource) jSource);
		return null;
	}

	public IDBConnection getConnection(JenaDBModelSource js) {
		try {
			return new DBConnection(new KeepAliveConnection(js.getDataSource().getConnection()),js.getType());
		} catch (SQLException e) {
			log.error("",e);
			return null;
		}
	}

	private boolean hasIndex(NonExceptingConnection conn, String tableName, String indexName) {
		NonExceptingResultSet rs = conn.getMetaData().getIndexInfo(null,null,tableName,false,true);
		if (rs==null || !rs.next()) return false;
		for (;!rs.isAfterLast();rs.next()) if (indexName.equalsIgnoreCase(rs.getString(6))) return true;
        return false;
	}

	@SuppressWarnings("null")
	public ModelRDB fixDBModel(JenaDBModelSource js) {
		IDBConnection conn = getConnection(js);
		if (!conn.isFormatOK()) conn.getDriver().setLongObjectLength(conn.getDriver().getLongObjectLengthMax());
		NonExceptingConnection sqconn = null;
		try { sqconn = new NonExceptingConnection(conn.getConnection()); } catch (SQLException e) { log.error("SQL error while fixing Jena database",e); }
		NonExceptingStatement stmt = sqconn.createStatement();
		ModelRDB m = (ModelRDB) ModelFactory.createModelRDBMaker(conn).openModel(js.getModel(),false);
		NonExceptingResultSet rs = null;
		if (js.getType().equals("HSQL") || js.getType().equals("Derby")) // Jouni, 28.10.2008
			rs = sqconn.getMetaData().getColumns(sqconn.getCatalog(),null,"JENA_GRAPH","LASTMODIFIED");
		else rs = sqconn.getMetaData().getColumns(sqconn.getCatalog(),null,"jena_graph","LastModified");
		if (!js.getType().equals("Derby") &&!rs.next()) { // Jouni, 28.10.2008
			stmt.executeUpdate("alter table jena_graph add column LastModified TimeStamp");
			stmt.executeUpdate("update jena_graph set LastModified=CURRENT_TIMESTAMP");
		}
		rs = stmt.executeQuery("select id from jena_graph where Name='"+js.getModel()+"'");
		rs.next();
		String tableName;
		if (js.getType().equals("HSQL"))
			tableName="JENA_G"+rs.getInt(1)+"T1_STMT";
		else tableName="jena_g"+rs.getInt(1)+"t1_stmt";
		if (hasIndex(sqconn,tableName,tableName+"XO")) {
			if (js.getType().equals("HSQL")) {
				stmt.executeUpdate("drop index "+tableName+"XO");
			} else {
				stmt.executeUpdate("drop index "+tableName+"XO on " +tableName);
			}
		}
		if (!hasIndex(sqconn,tableName,tableName+"XPO"))
			stmt.executeUpdate("create index "+tableName+"XPO on "+tableName+" (prop,obj)");
		if (!hasIndex(sqconn,tableName,tableName+"XSP"))
			stmt.executeUpdate("create index "+tableName+"XSP on "+tableName+" (subj,prop)");
		if (!hasIndex(sqconn,tableName,tableName+"XOS"))
			stmt.executeUpdate("create index "+tableName+"XOS on "+tableName+" (obj,subj)");
		return m;
	}

	public String expandPrefix(Model model, String uri) {
		return expandPrefix(model.expandPrefix(uri));
	}

	public String expandPrefix(String uri) {
		return globalPrefixMap.expandPrefix(uri);
	}

	public Model getModel(String o) {
		if (models.get(o) instanceof JenaModelSource) return addModelFromSource((IJenaModelSource)models.get(o));
		return (Model)models.get(o);
	}

	@SuppressWarnings("unchecked")
	public void initialize() {
		globalPrefixMap.withDefaultMappings(PrefixMapping.Standard);
		if (eagerLoad) for (IJenaModelSource js : sources) addModelFromSource(js);
		else for (IJenaModelSource js : sources) models.put(js.getKey(),js);
//		for (Model m : models.values()) copyMappings(globalPrefixMap,m); // it's probably bad to add namespace definitions from loaded models to the global prefix
		/*new Timer().schedule(new TimerTask() {
			@Override

			public void run() {
				for (IJenaUrlModelSource js : updateWatch) {
					boolean reload=false;
					for (String url: js.getUrls())
						try {
							if (new URL(url).openConnection().getLastModified()>lastModified.get(js.getKey())) {
								reload=true;
								break;
							}
						} catch (MalformedURLException e) {
							log.error("Malformed URL " +url+" when checking last modification for Jena model "+js.getKey(),e);
						} catch (IOException e) {
							log.error("Read from URL "+url+" failed when checking last modification for Jena model "+js.getKey(),e);
						}
					if (reload) {
						Model m = (Model) models.get(js.getKey());
						for (IModelReloadEventListener l : eventListeners.get(m)) l.startReloadEvent(m, js.getKey());
						m.removeAll();
						readModelFromUrlSource(js);
						for (IModelReloadEventListener l : eventListeners.get(m)) l.endReloadEvent(m, js.getKey());
					}
				}
			}

		},1000*30,1000*30);*/
	}

	public long getLastModified(Model model) {
		if (!models.containsValue(model)) {
			log.error("Attempted to get last modified date for model not registered in JenaService");
			return -1;
		}
		return lastModified.get(models.getKey(model));
	}

	public long getLastModified(String modelKey) {
		if (!(models.get(modelKey) instanceof Model)) addModelFromSource((IJenaModelSource) models.get(modelKey));
		return lastModified.get(modelKey);
	}

	public long getFileLastModified(String modelKey) {
		return getFileLastModified(modelKey, null);
	}
	
	public long getFileLastModified(String modelKey, List<String> ignoreKeys) {
		if (models.get(modelKey) instanceof JenaModelCombinationSource) {
			long newestTimestamp = -1;
			for (String key : ((JenaModelCombinationSource) models.get(modelKey)).getKeys()) {
				if (!ignoreKeys.contains(key)) {
					long currentTimestamp = getLastModified(key);
					if (currentTimestamp > newestTimestamp)
						newestTimestamp = currentTimestamp;
				}
			}
			return newestTimestamp;
		}
		else if (models.get(modelKey) instanceof JenaComboTDBModelSource)
			return new JenaUrlModelSource(modelKey, ((JenaComboTDBModelSource) models.get(modelKey)).getUrls()).getLastModified();
		else
			return this.getLastModified(modelKey);
	}
	
	public long getFileLastModified(Model model) {
		return this.getFileLastModified(model, null);
	}
	
	public long getFileLastModified(Model model, List<String> ignoreKeys) {
		return this.getFileLastModified(models.getKey(model), ignoreKeys); 
	}
	
	public static String parseType(String name) {
		int index = name.lastIndexOf('.');
		if (index!=-1) {
			String type = name.substring(name.lastIndexOf('.')).toLowerCase();
			if (type.equals(".rdf")) return "RDF/XML";
			if (type.equals(".rdfs")) return "RDF/XML";
			if (type.equals(".xml")) return "RDF/XML";
			if (type.equals(".owl")) return "RDF/XML";
			if (type.equals(".ttl")) return "TURTLE";
			if (type.equals(".n3")) return "N3";
			if (type.equals(".nt")) return "N-TRIPLE";
		}
		log.warn("Couldn't deduce file type for file "+name);
		return null;
	}
	
	public Property p(String uri) {
		return ResourceFactory.createProperty(globalPrefixMap.expandPrefix(uri));
	}

	public Resource r(String uri) {
		return ResourceFactory.createResource(globalPrefixMap.expandPrefix(uri));
	}

	public void putModel(String o,Model m) {
//		copyMappings(m,globalPrefixMap); // it's probably bad to add namespace definitions from loaded models to the global prefix
		copyMappings(globalPrefixMap,m);
		models.put(o,m);
		lastModified.put(o,System.currentTimeMillis());
		m.register(new ModelLastModifiedUpdateListener(o));
	}

	// Hivemind parameter
	public void setSources(List<JenaModelSource> sources) {
		this.sources = sources;
	}

	// Hivemind parameter
	public void setPrefixMappings(Map<String,String> prefixMappings) {
		this.globalPrefixMap.setNsPrefixes(prefixMappings);
	}

	public void setNsPrefix(String prefix, String namespace) {
		globalPrefixMap.setNsPrefix(prefix, namespace);
		// for (Model m : models.values()) m.setNsPrefix(prefix,namespace); // better not mess with the model's namespaces
	}

	public String shortForm(String uri) {
		return globalPrefixMap.shortForm(uri);
	}

	class ModelLastModifiedUpdateListener extends ObjectListener {

		private Object modelKey;

		public ModelLastModifiedUpdateListener(Object modelKey) {
			this.modelKey=modelKey;
		}

		@Override
		public void added(Object o) {
			lastModified.put(modelKey,System.currentTimeMillis());
		}

		@Override
		public void removed(Object o) {
			lastModified.put(modelKey,System.currentTimeMillis());
		}
	}

	public boolean isEagerLoad() {
		return eagerLoad;
	}

	public void setEagerLoad(boolean eagerLoad) {
		this.eagerLoad = eagerLoad;
	}

	public void addEventListener(Model model, IModelReloadEventListener listener) {
		Collection<IModelReloadEventListener> c = eventListeners.get(model);
		if (c==null) {
			c = new ArrayList<IModelReloadEventListener>();
			eventListeners.put(model,c);
		}
		c.add(listener);
	}

	public void removeEventListener(Model model, IModelReloadEventListener listener) {
		eventListeners.get(model).remove(listener);
	}

	public String getKey(Model projectionModel) {
		return models.getKey(projectionModel);
	}

	public Collection<String> getSourceUrls(String o) {
		ArrayList<String> sourceUrls = new ArrayList<String>();
		
		for (JenaModelSource s : this.sources) {
			if (s.getKey().equals(o)) {
				if (s instanceof IJenaUrlModelSource)
					sourceUrls.addAll(((IJenaUrlModelSource)s).getUrls());
				break;
			}	
		}
		return sourceUrls;
	}
	
}