package fi.seco.semweb.infrastructure.prova;

import java.io.File;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.mandarax.kernel.ConstantTerm;
import org.mandarax.kernel.InferenceException;
import org.mandarax.kernel.Result;
import org.mandarax.kernel.VariableTerm;

import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.StmtIterator;

import fi.seco.semweb.util.iterator.AMappingIterator;
import fi.seco.semweb.util.iterator.IIterableIterator;
import fi.seco.semweb.util.map.AMappingMap;

import ws.prova.Communicator;
import ws.prova.kernel.ProvaResultSet;
import ws.prova.parser.ParsingException;
import ws.prova.util.NullWriter;

public class Prova {

	private Communicator prova;
	private int key = 0;
	private static Log log = LogFactory.getLog(Prova.class);
	
	public Communicator getCommunicator() {
		return prova;
	}

	public Prova(String name) {
		this(name,false);
	}
	
	public Prova(String name, String rulebase) {
		this(name,false);
		consult(rulebase);
	}

	public Prova(String name, BufferedReader rulebase) {
		this(name,false);
		consult(rulebase);
	}


	public Prova(String name, String rulebase,boolean debug) {
		this(name,debug);
		consult(rulebase);
	}

	public Prova(String name, BufferedReader rulebase,boolean debug) {
		this(name,debug);
		consult(rulebase);
	}
	
	public Prova(String name, boolean debug) {
		try {
			prova = new Communicator(name, null, new BufferedReader(new InputStreamReader(Prova.class.getResource("rdf-base.prova").openStream())), 0, Communicator.SYNC);
			// If classloader fails to find rdf-base.prova null is returned and results into NullPointerException when calling openStream
		} catch (NullPointerException en) {
			log.error("Couldn't find rdf-base.prova rulebase",en);
		} catch (IOException e) {
			log.error("Couldn't read base rdf prova rules from rdf-base.prova",e);
		}
		if (!debug) prova.setPrintWriter(NullWriter.getPrintWriter());
	}
	
	/*key = -1 on file not found*/
	public int consult(File file) {
		try { 
			BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(file),"UTF-8"));
			return this.consult(br);
		} catch(FileNotFoundException fex) {
			log.error("File didn't exist: "+file+" exception was "+fex.getMessage());
			return -1;
		} catch(Exception e) {
		log.error("ERROR!!! "+e.getMessage());
		return -1;
	}
	}
	
	public List<Integer> consult(List<String> rulefilenames) {
		List<Integer> keys = new ArrayList<Integer>();
		for(String filename: rulefilenames) {
			keys.add(this.consult(new File(filename)));	
		}
		return keys;
	}
	
	/**
	 * Read a string of rules.
	 * @param rulebase is a rule or a collection of rules in a string. This is not a filename.
	 * @return identifying key for this rulebase so that it can be unconsulted.
	 */
	public int consult(String rulebase) {
		try {
		  prova.consultSync(rulebase,Integer.toString(++key),new Object[]{});
		} catch (Exception e) {
			log.error("Problem consulting rulebase "+rulebase,e);
		}
		return key;
	}
	
	public int consult(BufferedReader in) {
		try {
			prova.consultSync(":- eval(consult(_0)).",Integer.toString(++key), new Object[]{in});
		} catch (ParsingException e) {
			log.error("Failed to parse prova "+in.toString(),e);
		} catch (Exception e) {
			log.error("Problem consulting rulebase "+in.toString(),e);
		}
		
		return key;
	}

	public void unConsult(int key) {
		try {
			prova.unconsultSync(Integer.toString(key));
		} catch (Exception e) {
			log.error("Problem unconsulting key "+key,e);
		}
	}

	public IIterableIterator<Map<String, Object>> solve(String rule) {
		return solve(rule,new Object[0]);
	}

	@SuppressWarnings("unchecked")
	public IIterableIterator<Map<String, Object>> solve(String rule, Object[] input) {
		try {
			ProvaResultSet rs = (ProvaResultSet)prova.consultSync(":- solve(with_handler("+rule+")).",Integer.toString(++key),input).get(0);
		    final Exception e = rs.getException();
		    if (e!=null) throw e;
			return new AMappingIterator<Result,Map<String,Object>>(rs.iterator()) {
				@Override
				public Map<String, Object> map(Result src) {
					try {
						return new AMappingMap<VariableTerm,ConstantTerm,String,Object>(src.getResults()) {
							@Override
							public String mapKey(VariableTerm key) {
								return key.getName();
							}
							@Override
							public Object mapValue(ConstantTerm value) {
								return value.getObject();
							}
						};
					} catch (InferenceException e) {
						throw new RuntimeException(e);
					}
				}
			};
		} catch (Exception e) {
			log.error("Error solving rule "+rule,e);
			return null;
		}
	}

	public void eval(String rule) {
		eval(rule, new Object[0]);
	}

	public void eval(String rule, Object[] input) {
		try {
			ProvaResultSet rs = (ProvaResultSet) prova.consultSync(":- eval(with_handler("+rule+")).",Integer.toString(++key),input).get(0);
		    final Exception e = rs.getException();
		    if (e!=null) throw e;
		} catch (Exception e) {
			log.error("Error evaluating rule "+rule,e);
		}
	}
	
	public static StmtIterator listStatementsForProperty(Property property, Model model) {
		return model.listStatements(null,property,(RDFNode)null);
	}
	
	public static StmtIterator listStatementsForObject(RDFNode object, Model model) {
		return model.listStatements(null,null,object);
	}
	
	public static void handleException(String msg,Exception e) {
		log.error(msg,e);
	}
		
}
