package fi.seco.semweb.service.indexing.item;

import java.io.IOException;
import java.sql.Connection;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.ReadLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.WriteLock;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.search.FieldCache;
import org.apache.lucene.search.MatchAllDocsQuery;

import fi.seco.semweb.util.iterator.primitive.AIntToObjectMappingIterator;
import fi.seco.semweb.util.lucene.LuceneDirectory;
import fi.seco.semweb.util.lucene.NonExceptingHits;
import fi.seco.semweb.util.lucene.NonExceptingIndexReader;
import fi.seco.semweb.util.lucene.NonExceptingIndexSearcher;
import fi.seco.semweb.util.lucene.NonExceptingIndexWriter;
import fi.seco.semweb.util.sql.NonExceptingConnection;
import fi.seco.semweb.util.sql.NonExceptingPreparedStatement;
import fi.seco.semweb.util.sql.NonExceptingResultSet;

public class ItemIndexService {

	@SuppressWarnings("unused")
	private static final Log log = LogFactory.getLog(ItemIndexService.class);

	private ReentrantReadWriteLock rwl = new ReentrantReadWriteLock();

	private int itemCount;
	private ArrayList<String> allItemUris;

	private Collection<ICacheEventListener> listeners = new ArrayList<ICacheEventListener>();

	private NonExceptingConnection sqlConn;

	private LuceneDirectory directory;

	private Set<IItemLuceneIndexer> luceneIndexers = new HashSet<IItemLuceneIndexer>();
	private Set<IItemIndependentIndexer> independentIndexers = new HashSet<IItemIndependentIndexer>();

	public ItemIndexService() {
	}

	public ItemIndexService(String path, Connection conn) {
		this(path);
		setSqlConn(conn);
	}

	public ItemIndexService(String path) {
		this();
		setLucenePath(path);
	}

	public void setSqlConn(NonExceptingConnection sqlConn) {
		this.sqlConn=sqlConn;
	}

	public void setSqlConn(Connection sqlConn) {
		this.sqlConn=new NonExceptingConnection(sqlConn);
	}

	public void setLucenePath(String path) {
		directory = new LuceneDirectory(path);
		NonExceptingHits hits = directory.getLuceneSearcher().search(new MatchAllDocsQuery());
		allItemUris = new ArrayList<String>(hits.length()+1);
		allItemUris.add(null);
		String[] uris;
		try {
			uris = FieldCache.DEFAULT.getStrings(directory.getLuceneReader().getReader(),"uri");
			itemCount=uris.length;
			for (int i=0;i<hits.length();i++) allItemUris.add(uris[hits.id(i)]);
		} catch (IOException e) {
			log.error("Error reading Lucene index",e);
		}
	}

	public void setListener(ICacheEventListener listener) {
		this.listeners.add(listener);
	}

	public void setLuceneIndexer(IItemLuceneIndexer li) {
		luceneIndexers.add(li);
	}

	public boolean removeLuceneIndexer(IItemLuceneIndexer li) {
		return luceneIndexers.remove(li);
	}

	public void setIndependentIndexer(IItemIndependentIndexer li) {
		independentIndexers.add(li);
	}

	public boolean removeIndependentIndexer(IItemIndependentIndexer li) {
		return independentIndexers.remove(li);
	}

	public ReadLock readLock() {
		return rwl.readLock();
	}

	public WriteLock writeLock() {
		return rwl.writeLock();
	}

	public int getItemCount() {
		return itemCount;
	}

	public ArrayList<String> getAllItemUris() {
		return allItemUris;
	}

	public long getLastModified() {
		return directory.getLastModified();
	}

	public NonExceptingConnection getSqlConn() {
		return sqlConn;
	}

	public int addItems(List<String> resourceUris) {
		int ret = itemCount;
		rwl.writeLock().lock();
		log.debug("Adding "+resourceUris.size()+" items to indices.");
/*		if (log.isDebugEnabled()) {
			log.debug("Adding the following resources to the indices:");
			for (String s: resourceUris) { log.debug(" * "+s); }
		} */
		log.debug(" * Lucene index");
		addToLuceneIndex(resourceUris);
		log.debug(" done.");
		log.debug(" * SQL index");
		addToSQLIndex(itemCount, resourceUris);
		log.debug(" done.");
		for (IItemIndependentIndexer ni : independentIndexers) {
			log.debug(" * Index "+ni);
			ni.addToItemIndex(itemCount, resourceUris);
			log.debug(" done.");
		}
		log.debug(" * Memory cache of item id/uri's");
		for (String uri : resourceUris) allItemUris.add(uri);
		log.debug(" done.");
		for (ICacheEventListener il : listeners) il.invalidateCaches();
		rwl.writeLock().unlock();
		log.debug("done.");
		return ret;
	}

	public void removeItems(int[] ids) {
		rwl.writeLock().lock();
		removeFromLuceneIndex(ids);
		removeFromSQLIndex(ids);
		for (IItemIndependentIndexer ni : independentIndexers) ni.removeFromItemIndex(ids);
		for (int id : ids) allItemUris.remove(id);
		itemCount-=ids.length;
		for (ICacheEventListener il : listeners) il.invalidateCaches();
		rwl.writeLock().unlock();
	}

	public void clearIndices() {
		rwl.writeLock().lock();
		clearLuceneIndex();
		clearSQLIndex();
		for (IItemIndependentIndexer ni : independentIndexers) ni.clearItemIndex();
		itemCount=0;
		allItemUris.clear();
		allItemUris.add(null);
		for (ICacheEventListener il : listeners) il.invalidateCaches();
		rwl.writeLock().unlock();
	}

	public void clearIndexers() {
		independentIndexers.clear();
		luceneIndexers.clear();
	}
	public String mapFromIdToUri(int id) {
		return allItemUris.get(id);
	}

	public Iterable<String> mapFromIdsToUris(int[] ids) {
		return new AIntToObjectMappingIterator<String>(ids) {
			@Override
			public String map(int l) {
				return allItemUris.get(l);
			}

		};
	}

	private void clearLuceneIndex() {
		directory.clearIndex();
	}

	private void clearSQLIndex() {
		NonExceptingResultSet rs = sqlConn.getMetaData().getTables(sqlConn.getCatalog(),null,"ALL_ITEMS",null);
		if (rs.next()) sqlConn.createStatement().executeUpdate("DELETE FROM all_items");
	}

	private void removeFromLuceneIndex(int[] ids) {
		NonExceptingIndexReader reader = directory.getLuceneReader();
		for (int id : ids) reader.deleteDocument(id);
	}

	private void luceneIndexTruncatedAtIds(int[] ids) {
		itemCount-=ids.length;
		sqlLuceneIndexTruncatedAtIds(ids);
		for (IItemIndependentIndexer ni : independentIndexers) ni.truncateItemIndexAt(ids);
		for (int id : ids) allItemUris.remove(id);
	}

	private void sqlLuceneIndexTruncatedAtIds(int[] ids) {
		if (ids.length==1) sqlConn.createStatement().executeUpdate("UPDATE all_items SET id=id-1 WHERE id>"+ids[0]);
		else {
			int i=0;
			int lastId = ids[i++];
			while (i<ids.length-1) {
				sqlConn.createStatement().executeUpdate("UPDATE all_items SET id=id-"+i+" WHERE id>"+lastId+" AND id<"+ids[i++]);
				lastId=ids[i];
			}
			sqlConn.createStatement().executeUpdate("UPDATE all_items SET id=id-"+i+" WHERE id>"+lastId);
		}
	}

	private void addToLuceneIndex(Collection<String> resourceUris) {
		NonExceptingIndexWriter writer = directory.getLuceneWriter();
		for (String resourceUri : resourceUris) {
			Document d = new Document();
			for (IItemLuceneIndexer li : luceneIndexers) {
				Collection<Field> fields = li.getIndexFields(resourceUri,++itemCount);
				for (Field f : fields) d.add(f);
			}
			d.add(new Field("uri",resourceUri,Field.Store.YES,Field.Index.UN_TOKENIZED));
			writer.addDocument(d);
		}
		writer.optimize();
		//int[] t = writer.truncated();
		//if (t.length!=0) luceneIndexTruncatedAtIds(t);
	}

	private void addToSQLIndex(int maxItems, List<String> resourceUris) {
		NonExceptingResultSet rs = sqlConn.getMetaData().getTables(sqlConn.getCatalog(),null,"ALL_ITEMS",null);
		if (!rs.next()) {
			sqlConn.createStatement().executeUpdate("CREATE TABLE all_items (id INT, uri VARCHAR)");
			sqlConn.createStatement().executeUpdate("CREATE INDEX all_items_id_index ON all_items (id)");
			sqlConn.createStatement().executeUpdate("CREATE INDEX all_items_uri_index ON all_items (uri)");
		}
		NonExceptingPreparedStatement st = sqlConn.prepareStatement("INSERT INTO all_items (id,uri) VALUES (?,?)");
		for (String uri : resourceUris) {
			st.setInt(1, maxItems++);
			st.setString(2,uri);
			st.addBatch();
		}
		st.executeBatch();
	}

	private void removeFromSQLIndex(int[] ids) {
		NonExceptingPreparedStatement st = sqlConn.prepareStatement("DELETE FROM  all_items WHERE id=?");
		for (int id : ids) {
			st.setInt(0, id);
			st.addBatch();
		}
		st.executeBatch();
	}

	public NonExceptingIndexReader getLuceneReader() {
		return directory.getLuceneReader();
	}

	public NonExceptingIndexSearcher getLuceneSearcher() {
		return directory.getLuceneSearcher();
	}

}
