package fi.seco.semweb.service.lucenegator;

import java.io.IOException;

import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.Term;
import org.apache.lucene.index.TermDocs;
import org.apache.lucene.search.Explanation;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.Scorer;
import org.apache.lucene.search.Searcher;
import org.apache.lucene.search.Similarity;
import org.apache.lucene.search.TermQuery;
import org.apache.lucene.search.Weight;

/**
 * A simplified version of TermQuery to be used on categorical fields, i.e. fields where the
 * amount of text (fieldNorm) or the relative frequency of terms (idf) has no meaning.
 * A linear tf function is used (in fact it is the identity function, i.e., tf(freq)=freq) and
 * idf is not used as a factor in the calculation. However, query boosts and stored fieldNorms
 * are supported as usual. 
 * @author oisuomin
 *
 */

public class CategoryQuery extends TermQuery {

	private static final long serialVersionUID = 1L;

	/**
	 * A simplified version of TermWeight used by CategoryQuery. Most of the code is
	 * borrowed from TermWeight.
	 * @author oisuomin
	 *
	 */
	
	private class CategoryWeight implements Weight {

		private static final long serialVersionUID = 1L;

		private Similarity similarity;

		private float queryWeight;

		private float queryNorm;

		private float value;

		public CategoryWeight(Searcher searcher) {
			this.similarity = getSimilarity(searcher);
		}

		public Query getQuery() {
			return CategoryQuery.this;
		}

		public float getValue() {
			return value;
		}

		public float sumOfSquaredWeights() throws IOException {
			queryWeight = getBoost();
			return queryWeight * queryWeight;
		}

		public void normalize(float norm) {
			queryNorm = norm;
			queryWeight *= queryNorm;
			value = queryWeight;
		}

		public Scorer scorer(IndexReader reader) throws IOException {
			TermDocs termDocs = reader.termDocs(getTerm());

			if (termDocs == null)
				return null;

			return new CategoryScorer(this, termDocs, similarity, reader
					.norms(getTerm().field()));
		}

		public Explanation explain(IndexReader reader, int doc)
				throws IOException {
			Explanation result = new Explanation();
			result.setDescription("weight(" + getQuery() + " in " + doc
					+ "), product of:");

			Explanation queryExpl = new Explanation();
			queryExpl.setDescription("queryWeight(" + getQuery()
					+ "), product of:");

			Explanation boostExpl = new Explanation(getBoost(), "boost");
			if (getBoost() != 1.0f)
				queryExpl.addDetail(boostExpl);

			Explanation queryNormExpl = new Explanation(queryNorm, "queryNorm");
			queryExpl.addDetail(queryNormExpl);

			queryExpl.setValue(boostExpl.getValue() * queryNormExpl.getValue());

			result.addDetail(queryExpl);

			String field = getTerm().field();
			Explanation fieldExpl = new Explanation();
			fieldExpl.setDescription("fieldWeight(" + getTerm() + " in " + doc
					+ "), product of:");

			Explanation tfExpl = scorer(reader).explain(doc);
			fieldExpl.addDetail(tfExpl);

			Explanation fieldNormExpl = new Explanation();
			byte[] fieldNorms = reader.norms(field);
			float fieldNorm = fieldNorms != null ? Similarity
					.decodeNorm(fieldNorms[doc]) : 0.0f;
			fieldNormExpl.setValue(fieldNorm);
			fieldNormExpl.setDescription("fieldNorm(field=" + field + ", doc="
					+ doc + ")");
			fieldExpl.addDetail(fieldNormExpl);

			fieldExpl.setValue(tfExpl.getValue() * fieldNormExpl.getValue());

			result.addDetail(fieldExpl);

			result.setValue(queryExpl.getValue() * fieldExpl.getValue());

			if (queryExpl.getValue() == 1.0f)
				return fieldExpl;

			return result;
		}

		@Override
		public String toString() {
			return "weight(" + CategoryQuery.this + ")";
		}

	}

	/**
	 * Create a new CategoryQuery, with the supplied Term holding the category id.
	 * @param t term containing the category id to look for
	 */
	
	public CategoryQuery(Term t) {
		super(t);
	}

	@Override
	protected Weight createWeight(Searcher searcher) throws IOException {
		return new CategoryWeight(searcher);
	}

	/** Returns true iff <code>o</code> is equal to this. */
	@Override
	public boolean equals(Object o) {
		if (!(o instanceof CategoryQuery))
			return false;
		CategoryQuery other = (CategoryQuery) o;
		return (this.getBoost() == other.getBoost())
				&& this.getTerm().equals(other.getTerm());
	}

	/** Returns a hash code value for this object. */
	@Override
	public int hashCode() {
		return Float.floatToIntBits(getBoost()) ^ getTerm().hashCode();
	}
	
}
