package fi.seco.semweb.service.lucenegator;

import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.search.Similarity;
import org.apache.lucene.search.SimilarityDelegator;

/**
 * A Similarity implementation that intercepts calls to lengthNorm for a certain field,
 * returning a neutral value (1.0). For other fields and methods other than lengthNorm,
 * CategorySimilatiry delegates to another Similarity implementation. This version of
 * Similarity is used to ensure that the fieldNorm of category fields doesn't depend
 * on the length of the value. Thus, normally, fieldNorm == boost * lengthNorm but with
 * CategorySimilarity, fieldNorm == boost for the given field.
 * @author oisuomin
 *
 */

public class CategorySimilarity extends SimilarityDelegator {

	private static final long serialVersionUID = 1L;

	private String field;
	private Similarity delegee;

	/**
	 * Construct a new CategorySimilarity, delegating calls to the given Similarity
	 * except for lengthNorm calls for the given field name.
	 * @param delegee Similarity implementation to delegate to
	 * @param categoryFieldName name of field wose lengthNorm calls should be intercepted
	 */

	public CategorySimilarity(Similarity delegee, String categoryFieldName) {
		super(delegee);
		this.delegee = delegee;
		field = categoryFieldName;
	}

	@Override
	public float lengthNorm(String fieldName, int numTerms) {
		if (fieldName.equals(field)) // if the queried field name matches the defined categoryFieldName,
			return 1.0f; // return a neutral value
		// otherwise, delegate to the delegee
		return delegee.lengthNorm(fieldName, numTerms);
	}

	/**
	 * Convenience method to replace the Similarity implementation of an IndexWriter with
	 * a CategorySimilarity instance that delegates to the original Similarity except when
	 * asked for a fieldNorm for a pre-defined field.
	 * @param writer IndexWriter whose Similarity to replace
	 * @param categoryFieldName name of the field whose fieldNorm to intercept
	 */

	public static void install(IndexWriter writer, String categoryFieldName) {
		Similarity orig = writer.getSimilarity();
		CategorySimilarity cs = new CategorySimilarity(orig, categoryFieldName);
		writer.setSimilarity(cs);
	}

}
