package fi.seco.semweb.service.multimedia;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.imageio.ImageIO;
import javax.imageio.ImageWriter;
import javax.imageio.stream.ImageOutputStream;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import fi.seco.semweb.service.multimedia.model.CachingImageSpec;
import fi.seco.semweb.service.multimedia.model.IImageCachingSpec;
import fi.seco.semweb.service.multimedia.model.IImageMappingSpec;
import fi.seco.semweb.service.multimedia.model.IImageScalingSpec;
import fi.seco.semweb.service.multimedia.model.IImageSpecHolder;

/**
 * Renders previews of images and video.
 *
 * @author thaggstr
 *
 */
public class ImageService {

	private Log log = LogFactory.getLog(ImageService.class);
	private List<IImageSpecHolder> imageSpecHolders;

	private static final String DEFAULT_CACHE_LOCATION = "/common/group/finnonto/kulttuuriSampo/kuvat/tmpImgCache/";

	/**
	 * Creates an image according to provided spec from the image given as
	 * imageUri parameter.
	 *
	 * @param imageUri the path to the image
	 * @param spec the IImageSpec according to which to adjust the image
	 * @return the BufferedImage confirming to the IImageSpec
	 * @throws IOException
	 */
	/*public BufferedImage createImage(String imageUri, IImageSpec spec) throws IOException {
		log.debug("Creating image of " + imageUri + " according to spec " + spec);
		if(isCached(imageUri, spec)){
			return getCachedImage(imageUri, spec);
		} else {
			ImageScaler scaler = new ImageScaler(spec);
			imageUri = this.mapUri(imageUri);
			File f = new File(imageUri);
			BufferedImage origImage;
			try {
				origImage = ImageIO.read(f);
				if(origImage != null){
					log.debug("Image found and read. Proceeding to scale.");
					BufferedImage scaledImage = scaler.scale(origImage);
					return scaledImage;
				} else {
					log.error("No suitable readers could be found for " + imageUri);
					throw new IOException();
				}
			} catch (IOException e) {
				log.warn("Problem with image. " + imageUri + ".Problem: " + e.getMessage());
				throw e;
			}
		}
	}*/

	/**
	 * Creates an image according to found spec based on imageUri mapping.
	 * Searches for IImageSpecs that map to the given pattern
	 * and uses found IImageSpec to adjust the image.
	 *
	 * @param imageUri the application path to the image
	 * @return the BufferedImage confirming to the IImageSpec or null if an image can't be found
	 */
	public BufferedImage createImage(String imageUri) throws IOException {
		log.debug("Creating image of " + imageUri);
		// get the transformation specs, so we know what to do with the image
		// TODO a more general approach
		IImageScalingSpec scalingSpec = this.getScalingSpec(imageUri);
		if(scalingSpec != null){
			// based on the specs create cache name and check if it allready exists
			if(isCached(imageUri,scalingSpec))
				return getCachedImage(imageUri, scalingSpec);
			// otherwise create new image and save it in cache (specified or default)
			ImageScaler scaler = new ImageScaler(scalingSpec);
			String mappedImageUri = this.mapUri(imageUri);
			File f = new File(mappedImageUri);
			BufferedImage origImage;
			try {

				log.debug("Reading image "+f);
				origImage = ImageIO.read(f);
				if(origImage != null){
					log.debug("Image found and read. Proceeding to scale.");
					BufferedImage scaledImage = scaler.scale(origImage);
					log.debug("Image scaled");
					cacheImage(imageUri, scalingSpec, scaledImage);
					log.debug("Flushing original");
					origImage.flush();
					log.debug("Original flushed");
					return scaledImage;
				}
				log.error("ImageIO read returned null");
			} catch (IOException e) {
				log.warn("Problem with image. " + mappedImageUri + ".Problem: " + e.getMessage());
			} finally {
			}
		} else log.warn("Scaling spec not found!");
		return null;
	}

	private IImageScalingSpec getScalingSpec(String uri) {
		IImageScalingSpec result = null;
		boolean found = false;
		for(IImageSpecHolder holder:imageSpecHolders){
			IImageMappingSpec mapping = holder.getMappingSpec();
			if(mapping != null){
				String INPUT = uri;
				String REGEX = mapping.getPattern();
				if(REGEX != null){
					Pattern pattern;
					Matcher matcher;
					pattern = Pattern.compile(REGEX);
					matcher = pattern.matcher(INPUT);
					if(matcher.find()) if(!found){
						IImageScalingSpec tmp = holder.getScalingSpec();
						if(tmp != null){
							result = tmp;
							found = true;
						}
					} else log.warn("Multiple IImageCachingSpec matches for " + INPUT + ". Using the first one found.");
				}
			}
		}
		return result;
	}

	private IImageCachingSpec getCachingSpec(String uri) {
		IImageCachingSpec result = new CachingImageSpec();
		result.setCache(ImageService.DEFAULT_CACHE_LOCATION);
		boolean found = false;
		for(IImageSpecHolder holder:imageSpecHolders){
			IImageMappingSpec mapping = holder.getMappingSpec();
			if(mapping != null){
				String INPUT = uri;
				String REGEX = mapping.getPattern();
				Pattern pattern;
				Matcher matcher;
				pattern = Pattern.compile(REGEX);
				matcher = pattern.matcher(INPUT);
				if(matcher.find()) if(!found){
					IImageCachingSpec tmp = holder.getCachingSpec();
					if(tmp != null){
						result = tmp;
						found = true;
					}
				} else log.warn("Multiple IImageCachingSpec matches for " + INPUT + ". Using the first one found.");
			}
		}
		return result;
	}

	private String mapUri(String uri){
		boolean found = false;
		String INPUT = uri;
		String newURI = uri;
		for(IImageSpecHolder holder:imageSpecHolders){
			IImageMappingSpec mapping = holder.getMappingSpec();
			if(mapping != null){
				String REGEX = mapping.getPattern();
				String REPLACEMENT = mapping.getSrc();
				Pattern pattern;
				Matcher matcher;
				pattern = Pattern.compile(REGEX);
		        matcher = pattern.matcher(INPUT);
		        if(matcher.find()) if(!found){
					newURI = matcher.replaceFirst(REPLACEMENT);
					found = true;
				} else log.warn("Multiple mapping matches for " + INPUT + ". Using the first one found.");
			}
		}
		return newURI;
	}

	/**
	 * Check if the image given as parameter translated to
	 * a cache file name exists.
	 *
	 * @param imageUri the original application uri of the image
	 * @return
	 */
	private boolean isCached(String imageUri, IImageScalingSpec scalingSpec){
		File cacheFile = new File(createCacheFileName(imageUri,scalingSpec));
		log.debug("Checking if " + imageUri + " exists. Result: " + cacheFile.exists());
		return cacheFile.exists();
	}

	private BufferedImage getCachedImage(String imageUri, IImageScalingSpec scalingSpec) throws IOException {
		File cacheFile = new File(createCacheFileName(imageUri,scalingSpec));
		BufferedImage cachedImage;
		try {
			cachedImage = ImageIO.read(cacheFile);
			if(cachedImage != null){
				log.debug("Image " + imageUri + " found in cache and read.");
				return cachedImage;
			}
			log.error("No suitable readers could be found for " + imageUri);
			throw new IOException();
		} catch (IOException e) {
			log.warn("Problem with image. " + imageUri + ".Problem: " + e.getMessage());
			throw e;
		}
	}

	@SuppressWarnings("null")
	private void cacheImage(String imageUri, IImageScalingSpec scalingSpec, BufferedImage image) {
		String cacheFileName = createCacheFileName(imageUri,scalingSpec);
		log.debug("Caching version of image " + imageUri + " into " + cacheFileName);
		File cacheFile=new File(cacheFileName);
		boolean success;
		try {
			success = cacheFile.createNewFile();
			if (success) log.debug("File " + cacheFile.toString() + " did not exist and was created");
			else log.warn("File " + cacheFile.toString() + " already exists and will be overwritten!");
		} catch (IOException e) {
			log.error("Error creating cache file!", e);
		}
		FileOutputStream output = null;
		try {
			output = new FileOutputStream(cacheFile);
			log.debug("Getting usable writer");
			Iterator<ImageWriter> writers = ImageIO.getImageWritersByFormatName("jpg");
			ImageWriter writer = writers.next();
			log.debug("Usable writer gotten");
			// Once an ImageWriter has been obtained, its destination must be set to an ImageOutputStream:
			ImageOutputStream ios = null;
			try {
				ios = ImageIO.createImageOutputStream(output);
				writer.setOutput(ios);
				log.debug("Beginning write operation");
				writer.write(image);
				log.debug("Write operation finished");
				//output.close();
				//ios.close();
			} catch (IOException e) {
				log.error("",e);
			} finally {
				try {
					log.debug("Closing imagestream");
					ios.close();
					log.debug("Closed imagestream");
				} catch (IOException e) {
					log.error("Error closing ImageOutputStream! ", e);
				}
			}
		} catch (FileNotFoundException fnfe) {
			log.error("", fnfe);
		} finally {
			try {
				log.debug("Closing filestream");
				output.close();
				log.debug("Closed filestream");
			} catch (IOException e) {
				log.error("Error closing FileOutputStream! ", e);
			}
		}
		log.debug("Exiting cacheImage");
	}

	/**
	 * Takes a filename as first parameter for which it tries to locate
	 * a cache location. If none can be foud default is used. Uses
	 * the scalingSpec to generate the filename.
	 *
	 * @param originalImageFile
	 * @param scalingSpec
	 * @return
	 */
	private String createCacheFileName(String originalImageFile, IImageScalingSpec scalingSpec){
		String cacheLocation = "";
		// search for an IImageCachingSpec for the file
		IImageCachingSpec cachingSpec = this.getCachingSpec(originalImageFile);
		if(cachingSpec != null){
			cacheLocation = cachingSpec.getCache();
			log.debug("Caching spec found for image " + originalImageFile + " at " + cacheLocation);
		} else {
			log.warn("No caching spec found. Default location will be used.");
			cacheLocation = ImageService.DEFAULT_CACHE_LOCATION;
		}
		File originalImageFileFile  = new File(originalImageFile); // Helper for getting the name of the file
		return cacheLocation + scalingSpec.getWidth() + "x" + scalingSpec.getHeight() + "_" + scalingSpec.isAllowenlarging() + "_" + scalingSpec.isRetainaspect() + "_" + originalImageFileFile.getName();
	}

	public List<IImageSpecHolder> getImageSpecHolders() {
		return imageSpecHolders;
	}

	public void setImageSpecHolders(List<IImageSpecHolder> imageSpecHolders) {
		this.imageSpecHolders = imageSpecHolders;
	}

	/*public List<MappingImageSpec> getImageMappingSpec() {
		return imageMappingSpec;
	}

	public void setImageMappingSpec(List<MappingImageSpec> imageMappingSpec) {
		this.imageMappingSpec = imageMappingSpec;
	}*/
}
