package fi.seco.semweb.service.view.categorytree;

import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import fi.seco.semweb.service.indexing.item.IItemIndependentIndexer;
import fi.seco.semweb.service.indexing.tree.SearchableTreeIndexService;
import fi.seco.semweb.service.indexing.tree.model.IndexedNode;
import fi.seco.semweb.service.search.ResultSet;
import fi.seco.semweb.service.search.ItemSearchController;
import fi.seco.semweb.service.search.query.IntersectionQuery;
import fi.seco.semweb.util.sql.NonExceptingConnection;
import fi.seco.semweb.util.sql.NonExceptingPreparedStatement;
import fi.seco.semweb.util.sql.NonExceptingResultSet;
import gnu.trove.TIntArrayList;
import gnu.trove.TIntFloatHashMap;
import gnu.trove.TIntFloatIterator;

public class CategoryTreeIndexService<T> extends SearchableTreeIndexService<T> implements IItemIndependentIndexer {

	@SuppressWarnings("unused")
	private static Log log = LogFactory.getLog(CategoryTreeIndexService.class);

	private NonExceptingPreparedStatement addToItemTable;
	private NonExceptingPreparedStatement removeFromItemTable;
	private NonExceptingPreparedStatement truncateItemTable;

	private NonExceptingPreparedStatement getItemCount;
	private NonExceptingPreparedStatement getItems;

	private IItemLinker itemLinker;
	private ItemSearchController control;

	protected String itemTable;

	public String getItemTable() {
		return itemTable;
	}

	public CategoryTreeIndexService(String lucenePath, String tablePrefix, ItemSearchController control, NonExceptingConnection sqlconn) {
		super(lucenePath,tablePrefix,sqlconn);
		this.itemTable=tablePrefix+"_itemcategory";
		this.control=control;
		if (!sqlconn.getMetaData().getTables(sqlconn.getCatalog(),null,(itemTable).toUpperCase(),null).next()) {
			sqlconn.createStatement().executeUpdate("CREATE TABLE "+itemTable+" (id INT, cat INT, degree FLOAT)");
			sqlconn.createStatement().executeUpdate("CREATE INDEX "+itemTable+"_cat_index ON "+itemTable+" (cat)");
			sqlconn.createStatement().executeUpdate("CREATE INDEX "+itemTable+"_id_index ON "+itemTable+" (id)");
		}
		addToItemTable=sqlconn.prepareStatement("INSERT INTO "+itemTable+" (id, cat, degree) VALUES (?,?,?)");
		removeFromItemTable=sqlconn.prepareStatement("DELETE FROM "+itemTable+" WHERE id = ?");
		truncateItemTable=sqlconn.prepareStatement("UPDATE "+itemTable+" SET id=id-1 WHERE id>?");
		getItems=sqlconn.prepareStatement("SELECT id, MAX(degree) FROM "+itemTable+" WHERE cat>=? AND cat<=? GROUP BY id");
		getItemCount=sqlconn.prepareStatement("SELECT COUNT(DISTINCT(id)) FROM "+itemTable+" WHERE cat>=? AND cat<=?");
	}

	public CategoryTreeIndexService(String lucenePath, String tablePrefix, ItemSearchController control, NonExceptingConnection sqlconn, IItemLinker itemLinker) {
		this(lucenePath,tablePrefix,control,sqlconn);
		this.itemLinker=itemLinker;
	}

	public void setItemLinker(IItemLinker itemLinker) {
		this.itemLinker=itemLinker;
	}

	public void addToItemIndex(int startingId, List<String> resourceUris) {
		for (String uri : resourceUris) {
			addToItemTable.setInt(1,startingId);
			TIntFloatIterator tif = itemLinker.getLinkedCategories(this,startingId,uri).iterator();
			while (tif.hasNext()) {
				tif.advance();
				addToItemTable.setInt(2,tif.key());
				addToItemTable.setFloat(3,tif.value());
				addToItemTable.addBatch();
			}
			startingId++;
		}
		addToItemTable.executeBatch();
	}

	public void removeFromItemIndex(int[] ids) {
		for (int id : ids) {
			removeFromItemTable.setInt(1,id);
			removeFromItemTable.addBatch();
		}
		removeFromItemTable.executeBatch();
	}

	public void truncateItemIndexAt(int[] ids) {
		for (int id : ids) {
			truncateItemTable.setInt(1,id);
			truncateItemTable.addBatch();
		}
		truncateItemTable.executeBatch();
	}

	public void clearItemIndex() {
		sqlconn.createStatement().executeUpdate("DELETE FROM "+itemTable);
	}

	public TIntFloatHashMap getItems(IndexedNode<T> c) {
		TIntFloatHashMap ret = new TIntFloatHashMap();
		getItems.setInt(1,c.getNodeId());
		getItems.setInt(2,c.getNodeid_end());
		NonExceptingResultSet rs = getItems.executeQuery();
		while (rs.next()) ret.put(rs.getInt(1),rs.getFloat(2));
		return ret;
	}

	public int getItemCount(IndexedNode<T> c) {
		getItemCount.setInt(1,c.getNodeId());
		getItemCount.setInt(2,c.getNodeid_end());
		NonExceptingResultSet rs = getItemCount.executeQuery();
		rs.next();
		return rs.getInt(1);
	}

	@Override
	public String toString() {
		return "CategoryTreeIndexService with itemTable: "+itemTable;
	}

	public TIntFloatHashMap getFilteredItems(IndexedNode<T> c, ResultSet items) {
		if (items==null) return getItems(c);
		ResultSet rs = control.getResultSet(new IntersectionQuery().addToIntersection(items).addToIntersection(new CategoryTreeSelectorState(this,c)));
		TIntFloatHashMap ret = new TIntFloatHashMap();
		for (int id : rs.ids()) ret.put(id,(Float)rs.getInfo(id).get(itemTable+"_degree"));
		return ret;
/*		String tempTable = control.getResultAsSQLTempTable(items);
		NonExceptingResultSet rs = sqlconn.createStatement().executeQuery("SELECT id, MAX(degree) FROM "+itemTable+","+tempTable+" WHERE "+itemTable+".id="+tempTable+".id AND "+itemTable+".cat>="+c.getNodeId()+" AND "+itemTable+".cat<="+c.getNodeid_end()+" GROUP BY id");
		TIntFloatHashMap ret = new TIntFloatHashMap();
		while (rs.next()) ret.put(rs.getInt(1),rs.getFloat(2));
		return ret; */
	}

	public int getFilteredItemCount(IndexedNode<T> c, ResultSet items)  {
		if (items==null) return getItemCount(c);
		String tempTable = control.getResultAsSQLTempTable(items);
		NonExceptingResultSet rs = sqlconn.createStatement().executeQuery("SELECT COUNT(DISTINCT(id)) FROM "+itemTable+","+tempTable+" WHERE "+tempTable+".id="+itemTable+".id AND "+itemTable+".cat>="+c.getNodeId()+" AND "+itemTable+".cat<="+c.getNodeid_end());
		rs.next();
		return rs.getInt(1);
	}

	public TIntFloatHashMap getFilteredItemsDirect(IndexedNode<T> c, ResultSet items) {
		if (items==null) return getItems(c);
		String tempTable = control.getResultAsSQLTempTable(items);
		NonExceptingResultSet rs = sqlconn.createStatement().executeQuery("SELECT id, MAX(degree) FROM "+itemTable+","+tempTable+" WHERE "+itemTable+".id="+tempTable+".id AND "+itemTable+".cat="+c.getNodeId()+" GROUP BY id");
		TIntFloatHashMap ret = new TIntFloatHashMap();
		while (rs.next()) ret.put(rs.getInt(1),rs.getFloat(2));
		return ret;
	}

	public int getFilteredItemCountDirect(IndexedNode<T> c, ResultSet items)  {
		if (items==null) return getItemCount(c);
		String tempTable = control.getResultAsSQLTempTable(items);
		NonExceptingResultSet rs = sqlconn.createStatement().executeQuery("SELECT COUNT(DISTINCT(id)) FROM "+itemTable+","+tempTable+" WHERE "+tempTable+".id="+itemTable+".id AND "+itemTable+".cat="+c.getNodeId());
		rs.next();
		return rs.getInt(1);
	}

	public int getNotFilteredItemCount(IndexedNode<T> c, ResultSet items) {
		String tempTable;
		if (items==null) tempTable = "all_items";
		else tempTable = control.getResultAsSQLTempTable(items);
		NonExceptingResultSet rs = sqlconn.createStatement().executeQuery("SELECT COUNT(DISTINCT(id)) FROM "+itemTable+","+tempTable+" WHERE "+tempTable+".id="+itemTable+".id AND ("+itemTable+".cat<"+c.getNodeId()+" OR "+itemTable+".cat>"+c.getNodeid_end()+")");
		rs.next();
		return rs.getInt(1);
	}

	public int[] getNotFilteredItems(IndexedNode<T> c, ResultSet items) {
		String tempTable;
		TIntArrayList ret = new TIntArrayList();
		if (items==null) tempTable = "all_items";
		else tempTable = control.getResultAsSQLTempTable(items);
		NonExceptingResultSet rs = sqlconn.createStatement().executeQuery("SELECT DISTINCT(id) FROM "+itemTable+","+tempTable+" WHERE "+tempTable+".id="+itemTable+".id AND ("+itemTable+".cat<"+c.getNodeId()+" OR "+itemTable+".cat>"+c.getNodeid_end()+")");
		while (rs.next()) ret.add(rs.getInt(1));
		return ret.toNativeArray();
	}

}
